import sys

import numpy as np
import PIL.Image
from pyxpg import Filter, PrimitiveTopology

from ambra.config import CameraConfig, Config, GuiConfig
from ambra.lights import UniformEnvironmentLight
from ambra.materials import DiffuseMaterial, MaterialProperty, MaterialSamplingOptions
from ambra.primitives3d import Mesh
from ambra.property import ArrayImageProperty, UploadSettings
from ambra.utils import profile
from ambra.viewer import Viewer

data = np.asarray(PIL.Image.open(sys.argv[1]))
if data.shape[2] == 3:
    data = np.dstack((data, np.full(data.shape[:2], 255, data.dtype)))

test_sequence = False
if test_sequence:
    data = data[:256, :256]
    with profile.profile("Interpolate"):
        data = np.linspace(data, np.full_like(data, 255), 2000, dtype=np.uint8)
else:
    data = data[np.newaxis]

img_data = ArrayImageProperty(
    data, upload=UploadSettings(preupload=False, async_load=True, cpu_prefetch_count=2, gpu_prefetch_count=2)
)

positions = (
    np.array(
        [
            -1.0,
            -1.0,
            0.0,
            -1.0,
            1.0,
            0.0,
            1.0,
            -1.0,
            0.0,
            1.0,
            1.0,
            0.0,
        ]
    ).reshape((4, 3))
    * 10
)
uvs = np.array(
    [
        0.0,
        0.0,
        0.0,
        1.0,
        1.0,
        0.0,
        1.0,
        1.0,
    ]
).reshape((4, 2))

material = DiffuseMaterial(
    MaterialProperty(img_data, 3, srgb=True, mips=True, name="albedo"),
    sampling_options=MaterialSamplingOptions(mag_filter=Filter.NEAREST),
)
mesh = Mesh(
    positions,
    uvs=uvs,
    primitive_topology=PrimitiveTopology.TRIANGLE_STRIP,
    material=material,
)

l = UniformEnvironmentLight((0.5, 0.5, 0.5))

v = Viewer(
    config=Config(
        camera=CameraConfig(position=(0, 0, 1.0)),
        world_up=(0, 1, 0),
        gui=GuiConfig(stats=True, playback=True, renderer=True),
    )
)
v.scene.objects.extend([mesh, l])
v.run()



import numpy as np

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig, RendererConfig
from ambra.primitives3d import ColormapDistanceToPlane, ColormapKind, Grid, GridType, Voxels
from ambra.viewer import Viewer, imgui


def grid3d(x: np.ndarray, y: np.ndarray, z: np.ndarray) -> np.ndarray:
    x, y, z = np.meshgrid(x, y, z)
    return np.hstack((x.reshape(-1, 1), y.reshape(-1, 1), z.reshape(-1, 1)))


positions: np.ndarray = grid3d(
    np.linspace(-5, 5, 11, endpoint=True, dtype=np.float32),
    np.linspace(-5, 5, 11, endpoint=True, dtype=np.float32),
    np.linspace(-5, 5, 11, endpoint=True, dtype=np.float32),
)


class CustomViewer(Viewer):
    def on_gui(self):
        if imgui.begin("Voxels")[0]:
            u, border_size = imgui.drag_float("Border size", colors_voxels.border_size, 0.01, 0, 1.0)
            if u:
                colors_voxels.border_size = border_size
                uniform_color_voxels.border_size = border_size
                colormap_voxels.border_size = border_size
            u, border_factor = imgui.drag_float("Border factor", colors_voxels.border_factor, 0.01, 0.0, 2.0)
            if u:
                colors_voxels.border_factor = border_factor
                uniform_color_voxels.border_factor = border_factor
                colormap_voxels.border_factor = border_factor

        imgui.end()
        return super().on_gui()


viewer = CustomViewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        gui=GuiConfig(
            stats=True,
            inspector=True,
            renderer=True,
            playback=True,
        ),
        renderer=RendererConfig(
            msaa_samples=4,
        ),
        camera=CameraConfig(
            position=(0, 25, 25),
            z_near=0.1,
        ),
    ),
)

angle = np.linspace(0, 2.0 * np.pi, 100, dtype=np.float32)
sin = np.sin(angle * 0.5)
cos = np.cos(angle * 0.5)
rotation = np.column_stack((cos, np.zeros_like(sin), np.zeros_like(cos), sin))

colors: np.ndarray = grid3d(
    np.linspace(0, 1.0, 11, endpoint=True, dtype=np.float32),
    np.linspace(0, 1.0, 11, endpoint=True, dtype=np.float32),
    np.linspace(0, 1.0, 11, endpoint=True, dtype=np.float32),
)
colors = (
    (colors[:, 0] * 255).astype(np.uint32)
    | (colors[:, 1] * 255).astype(np.uint32) << 8
    | (colors[:, 2] * 255).astype(np.uint32) << 16
    | 0xFF000000
)
colors_voxels = Voxels(positions, 1.0, colors=colors, rotation=rotation, translation=(-17, 0, 0))
colormap_voxels = Voxels(positions, 1.0, colormap=ColormapDistanceToPlane(ColormapKind.JET, -5, 5), rotation=rotation)
uniform_color_voxels = Voxels(positions, 1.0, uniform_color=0xFFCCCCCC, rotation=rotation, translation=(17, 0, 0))

viewer.scene.objects.extend(
    [
        colors_voxels,
        colormap_voxels,
        uniform_color_voxels,
    ]
)
viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import logging
import os
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
from pyxpg import (
    AccelerationStructure,
    AccelerationStructureMesh,
    AccessFlags,
    AllocType,
    BorderColor,
    Buffer,
    BufferUsageFlags,
    CommandBuffer,
    CompareOp,
    ComputePipeline,
    DepthAttachment,
    DescriptorBindingFlags,
    DescriptorPool,
    DescriptorSet,
    DescriptorSetBinding,
    DescriptorSetLayout,
    DescriptorType,
    Device,
    DeviceFeatures,
    Filter,
    Format,
    Gui,
    Image,
    ImageAspectFlags,
    ImageBarrier,
    ImageLayout,
    ImageUsageFlags,
    ImageView,
    IndexType,
    LoadOp,
    MemoryBarrier,
    MemoryUsage,
    PhysicalDeviceType,
    PipelineStageFlags,
    PushConstantsRange,
    RenderingAttachment,
    ResolveMode,
    Sampler,
    SamplerAddressMode,
    SamplerMipmapMode,
    Shader,
    StoreOp,
    slang,
)

from .config import RendererConfig, RenderMode, UploadMethod
from .ffx import SPDPipeline
from .lights import (
    LIGHT_TYPES_INFO,
    DirectionalLight,
    EnvironmentLight,
    GGXLUTPipeline,
    GpuEnvironmentCubemaps,
    IBLParams,
    IBLPipeline,
    Light,
    LightTypes,
    UniformEnvironmentLight,
)
from .materials import (
    ColorMaterial,
    DiffuseMaterial,
    DiffuseSpecularMaterial,
    PBRMaterial,
)
from .property import BufferProperty, ImageProperty, Property, PropertyItem
from .renderer_frame import RendererFrame, SemaphoreInfo
from .scene import Object, Scene
from .shaders import compile
from .utils.descriptors import create_descriptor_layout_pool_and_set, create_descriptor_pool_and_sets
from .utils.gpu import (
    AccelerationStructureInstanceInfo,
    BufferUploadInfo,
    BulkUploader,
    ImageUploadInfo,
    UniformPool,
    UploadableBuffer,
    div_round_up,
    view_bytes,
)
from .utils.ring_buffer import RingBuffer
from .utils.threadpool import ThreadPool
from .viewport import PathTracerViewport, Viewport

if TYPE_CHECKING:
    from .gpu_property import GpuProperty
    from .materials import Material

logger = logging.getLogger(__name__)

SHADERS_PATH = Path(__file__).parent.joinpath("shaders")

MAX_WORKERS = 16
DEFAULT_WORKERS = 4
if sys.version_info >= (3, 13):

    def get_cpu_count() -> Optional[int]:
        return os.process_cpu_count()
else:

    def get_cpu_count() -> Optional[int]:
        return os.cpu_count()


@dataclass
class FrameInputs:
    image: Image
    command_buffer: CommandBuffer
    transfer_command_buffer: Optional[CommandBuffer]
    additional_semaphores: List[SemaphoreInfo]
    transfer_semaphores: List[SemaphoreInfo]


@dataclass
class PathTracer:
    # Static
    pipeline: ComputePipeline
    descriptor_set: DescriptorSet
    descriptor_pool: DescriptorPool
    descriptor_set_layout: DescriptorSetLayout

    viewport_descriptor_sets: List[DescriptorSet]
    viewport_descriptor_pool: DescriptorPool
    viewport_descriptor_set_layout: DescriptorSetLayout

    # Scene
    acceleration_structure: Optional[AccelerationStructure]
    instances_buf: Optional[Buffer]
    materials_buf: Optional[Buffer]


class Renderer:
    def __init__(
        self,
        device: Device,
        width: int,
        height: int,
        num_frames_in_flight: int,
        output_format: Format,
        multiviewport: bool,
        max_viewports: int,
        config: RendererConfig,
    ):
        self.device = device
        self.num_frames_in_flight = num_frames_in_flight
        self.output_format = output_format
        self.multiviewport = multiviewport
        self.max_viewports = max_viewports
        self.shadow_map_format = Format.D32_SFLOAT

        # Config
        self.background_color = config.background_color
        self.supports_path_tracing = bool(device.features & DeviceFeatures.RAY_QUERY)
        if config.render_mode == RenderMode.PATH_TRACER and not self.supports_path_tracing:
            logger.warning(
                "Path tracing cannot be enabled because DeviceFeatures.RAY_QUERY is not supported. Falling back to raster mode."
            )
            self.render_mode = RenderMode.RASTER
        else:
            self.render_mode = config.render_mode

        # Path tracing
        self.path_tracer: Optional[PathTracer] = None
        self.path_tracer_max_bounces = config.path_tracer_max_bounces
        self.path_tracer_samples_per_frame = config.path_tracer_samples_per_frame
        self.path_tracer_max_samples_per_pixel = config.path_tracer_max_samples_per_pixel
        self.path_tracer_max_textures = config.path_tracer_max_textures
        self.path_tracer_clip_value = config.path_tracer_clip_value
        self.path_tracer_use_background_color = config.path_tracer_use_background_color
        self.path_tracer_instance_dtype = np.dtype(
            {
                "normal_matrix": (np.dtype((np.float32, (3, 4))), 0),
                "positions": (np.uint64, 48),
                "normals": (np.uint64, 56),
                "tangents": (np.uint64, 64),
                "uvs": (np.uint64, 72),
                "indices": (np.uint64, 80),
                "material_index": (np.uint32, 88),
                "_padding": (np.uint32, 92),
            }
        )  # type: ignore
        self.path_tracer_material_dtype = np.dtype(
            {
                "albedo": (np.dtype((np.float32, (3,))), 0),
                "albedo_texture": (np.uint32, 12),
                "roughness": (np.float32, 16),
                "roughness_texture": (np.uint32, 20),
                "metallic": (np.float32, 24),
                "metallic_texture": (np.uint32, 28),
                "normal_texture": (np.uint32, 32),
                "_padding": (np.uint32, 36),
                "_padding1": (np.uint32, 40),
                "_padding2": (np.uint32, 44),
            }
        )  # type: ignore
        self.path_tracer_constants_dtype = np.dtype(
            {
                "width": (np.uint32, 0),
                "height": (np.uint32, 4),
                "max_bounces": (np.uint32, 8),
                "num_directional_lights": (np.uint32, 12),
                "camera_position": (np.dtype((np.float32, (3,))), 16),
                "viewport_mask": (np.uint32, 28),
                "camera_forward": (np.dtype((np.float32, (3,))), 32),
                "film_dist": (np.float32, 44),
                "camera_up": (np.dtype((np.float32, (3,))), 48),
                "max_samples_per_pixel": (np.uint32, 60),
                "camera_right": (np.dtype((np.float32, (3,))), 64),
                "clip_value": (np.float32, 76),
                "ambient_light": (np.dtype((np.float32, (3,))), 80),
                "has_environment_light": (np.uint32, 92),
                "background_color": (np.dtype((np.float32, (3,))), 96),
                "use_background_color": (np.uint32, 108),
            }
        )  # type: ignore
        self.path_tracer_push_constants_dtype = np.dtype(
            {
                "sample_index": (np.uint32, 0),
            }
        )  # type: ignore
        self.path_tracer_push_constants = np.zeros(1, self.path_tracer_push_constants_dtype)

        # Scene descriptors
        self.max_shadow_maps = config.max_shadow_maps

        self.linear_sampler = Sampler(
            device,
            Filter.LINEAR,
            Filter.LINEAR,
            SamplerMipmapMode.LINEAR,
            u=SamplerAddressMode.CLAMP_TO_EDGE,
            v=SamplerAddressMode.CLAMP_TO_EDGE,
            name="linear-sampler",
        )
        self.shadow_sampler = Sampler(
            device,
            Filter.LINEAR,
            Filter.LINEAR,
            u=SamplerAddressMode.CLAMP_TO_BORDER,
            v=SamplerAddressMode.CLAMP_TO_BORDER,
            compare_enable=True,
            compare_op=CompareOp.LESS,
            border_color=BorderColor.FLOAT_OPAQUE_WHITE,
            name="shadow-sampler",
        )

        self.zero_buffer = Buffer.from_data(
            device,
            np.zeros((4,), np.uint8),
            BufferUsageFlags.STORAGE | BufferUsageFlags.UNIFORM | BufferUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            name="zero-buffer",
        )
        self.zero_image = Image.from_data(
            device,
            np.zeros((1, 1, 4), np.uint8),
            ImageLayout.SHADER_READ_ONLY_OPTIMAL,
            1,
            1,
            Format.R8G8B8A8_UNORM,
            ImageUsageFlags.SAMPLED | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            name="zero-image",
        )
        self.zero_cubemap = Image(
            self.device,
            1,
            1,
            Format.R16G16B16A16_SFLOAT,
            ImageUsageFlags.SAMPLED | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            array_layers=6,
            is_cube=True,
        )
        self.zero_triangle: Optional[Buffer] = None
        self.zero_material: Optional[Material] = None

        with device.sync_commands() as cmd:
            cmd.image_barrier(
                self.zero_cubemap,
                ImageLayout.TRANSFER_DST_OPTIMAL,
                MemoryUsage.NONE,
                MemoryUsage.TRANSFER_DST,
                undefined=True,
            )
            cmd.clear_color_image(self.zero_cubemap, (0, 0, 0, 1))
            cmd.image_barrier(
                self.zero_cubemap,
                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                MemoryUsage.TRANSFER_DST,
                MemoryUsage.SHADER_READ_ONLY,
            )

        # Maps from the index of the frame starting from which it's safe to destroy objects to the list of objets to destroy.
        self.destruction_queue: Dict[int, List[Union[Buffer, Image, ImageView]]] = {}
        self.shader_cache: Dict[Tuple[Union[str, Tuple[str, str], Path], ...], slang.Shader] = {}

        self.scene_descriptor_set_layout = DescriptorSetLayout(
            device,
            [
                DescriptorSetBinding(1, DescriptorType.UNIFORM_BUFFER),  # 0 - Constants
                DescriptorSetBinding(1, DescriptorType.SAMPLER),  # 1 - shadow_map sampler
                DescriptorSetBinding(1, DescriptorType.SAMPLER),  # 2 - Cubemap sampler
                DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE),  # 3 - Environment irradiance cubemap
                DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE),  # 4 - Environment specular cubemap
                DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE),  # 5 - Environment GGX LUT
                *[DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER) for _ in LIGHT_TYPES_INFO],
                DescriptorSetBinding(self.max_shadow_maps, DescriptorType.SAMPLED_IMAGE),
            ],
            name="scene-descriptor-layout",
        )

        self.scene_depth_descriptor_set_layout = DescriptorSetLayout(
            device,
            [
                DescriptorSetBinding(1, DescriptorType.UNIFORM_BUFFER),
            ],
            name="scene-depth-descriptor-set-layout",
        )

        self.scene_constants_dtype = np.dtype(
            {
                "view": (np.dtype((np.float32, (4, 4))), 0),
                "projection": (np.dtype((np.float32, (4, 4))), 64),
                "world_camera_position": (np.dtype((np.float32, (3,))), 128),
                "ambient_light": (np.dtype((np.float32, (3,))), 144),
                "has_environment_light": (np.uint32, 156),
                "focal": (np.dtype((np.float32, (2,))), 160),
                "inverse_viewport_size": (np.dtype((np.float32, (2,))), 168),
                "max_specular_mip": (np.float32, 176),
                "num_lights": (np.dtype((np.uint32, (len(LIGHT_TYPES_INFO),))), 180),
            }
        )  # type: ignore

        self.max_lights_per_type = config.max_lights_per_type

        self.spd_pipeline = SPDPipeline(self)
        self.spd_pipeline_instances = [
            self.spd_pipeline.alloc_instance(self, True) for _ in range(config.mip_generation_batch_size)
        ]

        self.ibl_default_params = IBLParams()
        self.ibl_pipeline: Optional[IBLPipeline] = None
        self.ggx_lut_pipeline = GGXLUTPipeline(self)
        self.ggx_lut = self.ggx_lut_pipeline.run(self)

        self.uniform_pool = UniformPool(device, self.num_frames_in_flight, config.uniform_pool_block_size)

        if config.thread_pool_workers is not None:
            self.num_workers = config.thread_pool_workers
        else:
            num_cpus = get_cpu_count()
            if num_cpus is not None:
                self.num_workers = min(num_cpus, MAX_WORKERS)
            else:
                self.num_workers = DEFAULT_WORKERS

        self.thread_pool: ThreadPool = ThreadPool(self.num_workers)
        self.total_frame_index = 0

        # Upload method for buffers
        if config.force_buffer_upload_method is not None:
            self.buffer_upload_method = config.force_buffer_upload_method
            if self.buffer_upload_method == UploadMethod.TRANSFER_QUEUE:
                if not device.has_transfer_queue:
                    raise RuntimeError("Transfer queue not available on picked device")
                if not device.features & DeviceFeatures.TIMELINE_SEMAPHORES:
                    raise RuntimeError(
                        "Transfer queue upload requires timeline semaphores which are not available on picked device"
                    )
        else:
            # TODO: we could also check if BAR is available and large enough
            # and add a DEVICE_PREFER_MAPPED option to take advantage of it.
            if (
                device.device_properties.device_type == PhysicalDeviceType.INTEGRATED_GPU
                or device.device_properties.device_type == PhysicalDeviceType.CPU
            ):
                self.buffer_upload_method = UploadMethod.MAPPED_PREFER_DEVICE
            elif (
                config.use_transfer_queue_if_available
                and device.has_transfer_queue
                and device.features & DeviceFeatures.TIMELINE_SEMAPHORES
            ):
                self.buffer_upload_method = UploadMethod.TRANSFER_QUEUE
            else:
                self.buffer_upload_method = UploadMethod.GRAPHICS_QUEUE

        # Upload method for images
        if config.force_image_upload_method is not None:
            if (
                config.force_image_upload_method == UploadMethod.MAPPED_PREFER_DEVICE
                or config.force_image_upload_method == UploadMethod.MAPPED_PREFER_HOST
            ):
                raise RuntimeError(
                    f"Upload method for images must be {UploadMethod.GRAPHICS_QUEUE} or {UploadMethod.TRANSFER_QUEUE}. Got {config.force_image_upload_method} in config.force_image_upload_method."
                )
            self.image_upload_method = config.force_image_upload_method
            if self.image_upload_method == UploadMethod.TRANSFER_QUEUE:
                if not device.has_transfer_queue:
                    raise RuntimeError("Transfer queue not available on picked device")
                if not device.features & DeviceFeatures.TIMELINE_SEMAPHORES:
                    raise RuntimeError(
                        "Transfer queue upload requires timeline semaphores which are not available on picked device"
                    )
        else:
            if (
                config.use_transfer_queue_if_available
                and device.has_transfer_queue
                and device.features & DeviceFeatures.TIMELINE_SEMAPHORES
            ):
                self.image_upload_method = UploadMethod.TRANSFER_QUEUE
            else:
                self.image_upload_method = UploadMethod.GRAPHICS_QUEUE

        # Allocate buffers for bulk upload
        self.bulk_uploader = BulkUploader(self.device, config.upload_buffer_size, config.upload_buffer_count)
        self.bulk_upload_list: List[Union[BufferUploadInfo, ImageUploadInfo]] = []

        # Enabled gpu properties cache for prefetch after frame submission
        self.enabled_gpu_properties: Set[GpuProperty[Any]] = set()

        # MSAA
        # TODO:
        # - proper thing to do would be to check max supported count for the color
        #   and depth format in use.
        # - additionally we could use StoreOp.DONT_CARE, TRANSIENT_ATTACHMENT and LAZILY_ALLOCATED memory
        #   for best performance on tilers.
        # - if at some point we want to expose multiple MSAA passes we can add a way to disable it.
        # - for retrieving color/depth we can do it by returning the resolved framebuffer (allocating one more for depth).
        self.msaa_samples = max(1, config.msaa_samples)
        self.msaa_target: Optional[Image] = None

        # Will be populated in self.resize(), and will never be None after that
        self.depth_buffer: Image = None  # type: ignore
        self.depth_format = Format.D32_SFLOAT
        self.depth_clear_value = 1.0
        self.depth_compare_op = CompareOp.LESS
        self.resize(width, height)

    def resize(self, width: int, height: int) -> None:
        self.render_width = width
        self.render_height = height

        if self.depth_buffer is not None:
            self.depth_buffer.destroy()

        self.depth_buffer = Image(
            self.device,
            width,
            height,
            self.depth_format,
            ImageUsageFlags.DEPTH_STENCIL_ATTACHMENT | ImageUsageFlags.TRANSFER_DST,
            AllocType.DEVICE_DEDICATED,
            samples=self.msaa_samples,
            name="depth",
        )
        if self.msaa_samples > 1:
            if self.msaa_target is not None:
                self.msaa_target.destroy()
            self.msaa_target = Image(
                self.device,
                width,
                height,
                self.output_format,
                ImageUsageFlags.COLOR_ATTACHMENT,
                AllocType.DEVICE_DEDICATED,
                samples=self.msaa_samples,
            )

    def run_ibl_pipeline(
        self, equirectangular: Image, ibl_params: Optional["IBLParams"] = None
    ) -> "GpuEnvironmentCubemaps":
        if self.ibl_pipeline is None:
            self.ibl_pipeline = IBLPipeline(self)
        if ibl_params is None:
            ibl_params = self.ibl_default_params
        return self.ibl_pipeline.run(self, equirectangular, ibl_params)

    def compile_shader(
        self,
        path: Union[Path, str],
        entry: str = "main",
        target: str = "spirv_1_3",
        defines: Optional[List[Tuple[str, str]]] = None,
        include_paths: Optional[List[Union[Path, str]]] = None,
    ) -> slang.Shader:
        defines = defines or []
        include_paths = include_paths or []
        key = (path, entry, target, *defines, *include_paths)
        if s := self.shader_cache.get(key):
            return s
        shader = compile(Path(path), entry, target, defines=defines, include_paths=include_paths)
        self.shader_cache[key] = shader
        return shader

    def compile_builtin_shader(
        self,
        path: Union[Path, str],
        entry: str = "main",
        target: str = "spirv_1_3",
        defines: Optional[List[Tuple[str, str]]] = None,
        include_paths: Optional[List[Union[Path, str]]] = None,
    ) -> slang.Shader:
        return self.compile_shader(SHADERS_PATH.joinpath(SHADERS_PATH, path), entry, target, defines, include_paths)

    def render(self, scene: Scene, viewports: List[Viewport], frame: FrameInputs) -> None:
        # Stage: destroy resources enqueued for destruction for this or an earlier frame
        if self.destruction_queue:
            for index, resources in list(self.destruction_queue.items()):
                if index <= self.total_frame_index:
                    for resource in resources:
                        resource.destroy()
                del self.destruction_queue[index]

        # Stage: create new objects
        enabled_objects: List[Object] = []
        enabled_lights: List[Light] = []
        enabled_gpu_properties: Set[GpuProperty[Any]] = set()

        def visit(o: Object) -> None:
            # Filter objects that are hidden in the GUI or that are not enabled this frame
            if o.gui_enabled and o.enabled.get_current():
                properties_enabled = True
                o.create_if_needed(self)

                if o.material is not None:
                    for mp, _ in o.material.properties:
                        if mp.property.current_animation_enabled:
                            mp.property.create(self)
                            if mp.property.gpu_property is not None:
                                enabled_gpu_properties.add(mp.property.gpu_property)
                        else:
                            properties_enabled = False
                for p in o.properties:
                    if p.current_animation_enabled:
                        p.create(self)
                        if p.gpu_property is not None:
                            enabled_gpu_properties.add(p.gpu_property)
                    else:
                        properties_enabled = False

                # Filter objects that have any property disabled for this frame
                if properties_enabled:
                    if isinstance(o, Light):
                        enabled_lights.append(o)
                    enabled_objects.append(o)

        scene.visit_objects(visit)

        enabled_widgets = []
        for w in scene.widgets:
            properties_enabled = True
            w.create_if_needed(self)
            for wp in w.properties:
                if wp.current_animation_enabled:
                    wp.create(self)
                    if wp.gpu_property is not None:
                        enabled_gpu_properties.add(wp.gpu_property)
                else:
                    properties_enabled = False
            if properties_enabled:
                enabled_widgets.append(w)

        # Stage: synchronous buffer upload after creating new objects
        if len(self.bulk_upload_list) > 0:
            mip_generation_requests: List[ImageUploadInfo] = []
            self.bulk_uploader.bulk_upload(self.bulk_upload_list, mip_generation_requests)
            self.bulk_upload_list.clear()

            # Process mip generation requests in batches.
            # The reason why this happens with sync command batches is that we
            # don't know upfront how many mips we have to create and we are
            # not allocating a pipeline instance (buffers and descriptors) to
            # run each of these generations independently.
            batch_size = len(self.spd_pipeline_instances)
            for i in range(0, len(mip_generation_requests), batch_size):
                with self.device.sync_commands() as cmd:
                    for j, spd_instance in enumerate(self.spd_pipeline_instances):
                        if i + j >= len(mip_generation_requests):
                            break
                        req = mip_generation_requests[i + j]
                        assert req.level_0_view is not None

                        spd_instance.set_image_extents(req.image.width, req.image.height, req.image.mip_levels)
                        self.spd_pipeline.run(
                            cmd,
                            req.image,
                            req.layout,
                            req.level_0_view,
                            req.mip_views,
                            spd_instance,
                            req.mip_generation_filter,
                        )

        path_tracing = self.render_mode == RenderMode.PATH_TRACER

        # Stage: path tracing pipeline creation
        if path_tracing and self.path_tracer is None:
            path_tracer_descriptor_set_layout, path_tracer_descriptor_pool, path_tracer_descriptor_set = (
                create_descriptor_layout_pool_and_set(
                    self.device,
                    [
                        DescriptorSetBinding(1, DescriptorType.ACCELERATION_STRUCTURE),
                        DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),
                        DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),
                        DescriptorSetBinding(1, DescriptorType.SAMPLER),
                        DescriptorSetBinding(
                            self.path_tracer_max_textures,
                            DescriptorType.SAMPLED_IMAGE,
                            DescriptorBindingFlags.VARIABLE_DESCRIPTOR_COUNT | DescriptorBindingFlags.PARTIALLY_BOUND,
                        ),
                    ],
                    self.path_tracer_max_textures,
                    name="path-tracer",
                )
            )
            path_tracer_viewport_descriptor_set_layout = DescriptorSetLayout(
                self.device,
                [
                    DescriptorSetBinding(1, DescriptorType.UNIFORM_BUFFER),  # 0 - Constants
                    *[DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER) for _ in LIGHT_TYPES_INFO],  # 1 - Lights
                    DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),  # 2 - Output
                    DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE),  # 3 - Environment light
                    DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),  # 4 - Accumulation
                ],
                name="path-tracer-viewport-descriptor-layout",
            )
            shader = self.compile_builtin_shader(Path("3d", "path_tracer", "path_tracer.slang"))
            path_tracer_pipeline = ComputePipeline(
                self.device,
                Shader(self.device, shader.code),
                descriptor_set_layouts=[
                    path_tracer_descriptor_set_layout,
                    path_tracer_viewport_descriptor_set_layout,
                ],
                push_constants_ranges=[PushConstantsRange(self.path_tracer_push_constants_dtype.itemsize)],
                name="pathtracer",
            )

            # TODO: dynamic viewport creation
            viewport_descriptor_pool, viewport_descriptor_sets = create_descriptor_pool_and_sets(
                self.device,
                path_tracer_viewport_descriptor_set_layout,
                self.max_viewports * self.num_frames_in_flight,
                name="pathtracer-viewport-scene-descriptor-sets",
            )
            self.viewports: List[Viewport] = []
            for viewport_index, v in enumerate(viewports):
                descriptor_sets = RingBuffer(
                    viewport_descriptor_sets[
                        viewport_index * self.num_frames_in_flight : (viewport_index + 1) * self.num_frames_in_flight
                    ]
                )
                uniform_buffers = RingBuffer(
                    [
                        UploadableBuffer(
                            self.device, self.path_tracer_constants_dtype.itemsize, BufferUsageFlags.UNIFORM
                        )
                        for _ in range(self.num_frames_in_flight)
                    ]
                )
                constants = np.zeros((1,), self.path_tracer_constants_dtype)
                for s, buf, light_bufs in zip(descriptor_sets, uniform_buffers, v.scene_light_buffers):
                    s.write_buffer(buf, DescriptorType.UNIFORM_BUFFER, 0)
                    for i, light_buf in enumerate(light_bufs):
                        s.write_buffer(light_buf, DescriptorType.STORAGE_BUFFER, 1, i)
                v.path_tracer_viewport = PathTracerViewport(
                    descriptor_sets,
                    uniform_buffers,
                    constants,
                    0,
                    v.camera.camera_from_world.copy(),
                    None,
                )

            self.path_tracer = PathTracer(
                pipeline=path_tracer_pipeline,
                descriptor_set=path_tracer_descriptor_set,
                descriptor_pool=path_tracer_descriptor_pool,
                descriptor_set_layout=path_tracer_descriptor_set_layout,
                viewport_descriptor_set_layout=path_tracer_viewport_descriptor_set_layout,
                viewport_descriptor_pool=viewport_descriptor_pool,
                viewport_descriptor_sets=viewport_descriptor_sets,
                # Scene
                acceleration_structure=None,
                instances_buf=None,
                materials_buf=None,
            )

        cmd = frame.command_buffer
        f = RendererFrame(
            # Frame counters
            self.total_frame_index % self.num_frames_in_flight,
            self.total_frame_index,
            path_tracing,
            # Graphics command list
            cmd,
            # Upload stages and barriers
            PipelineStageFlags(0),
            [],
            [],
            [],
            [],
            # Mip Generation requests
            {},
            # Before render
            PipelineStageFlags(0),
            PipelineStageFlags(0),
            [],
            # Between viewport render
            PipelineStageFlags(0),
            PipelineStageFlags(0),
            # Frame additional semaphores
            frame.additional_semaphores,
            # Transfer queue commands
            frame.transfer_command_buffer,
            # Transfer queue semaphores
            frame.transfer_semaphores,
            # Transfer queue barriers
            [],
            [],
            [],
            [],
        )

        # Stage: Load properties in CPU memory and prepare upload_before barriers
        for p in enabled_gpu_properties:
            p.load(f)

        # Sync: before-upload barriers
        cmd.barriers(
            memory_barriers=[
                MemoryBarrier(
                    PipelineStageFlags.ALL_COMMANDS,  # Would need to store the last stage as well
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                    PipelineStageFlags.TRANSFER,
                    AccessFlags.TRANSFER_READ | AccessFlags.TRANSFER_WRITE,
                ),
            ],
            buffer_barriers=f.upload_before_buffer_barriers,
            image_barriers=f.upload_before_image_barriers,
        )

        # Sync: before-upload transfer queue barriers (layout transitions)
        if frame.transfer_semaphores:
            assert frame.transfer_command_buffer is not None
            frame.transfer_command_buffer.barriers(
                memory_barriers=[
                    MemoryBarrier(
                        PipelineStageFlags.ALL_COMMANDS,
                        AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                        PipelineStageFlags.TRANSFER,
                        AccessFlags.TRANSFER_READ | AccessFlags.TRANSFER_WRITE,
                    )
                ],
                buffer_barriers=f.transfer_upload_before_buffer_barriers,
                image_barriers=f.transfer_upload_before_image_barriers,
            )

        # Stage: upload all properties, objects and materials and colect upload_after barriers
        for p in enabled_gpu_properties:
            p.upload(f)

        # Upload scene constants for each viewport
        for viewport_index, viewport in enumerate(viewports):
            viewport_width, viewport_height = viewport.rect.width, viewport.rect.height

            if viewport_width <= 0 or viewport_height <= 0:
                continue

            environment_light: Optional[EnvironmentLight] = None
            uniform_environment_radiance = np.array((0, 0, 0), np.float32)
            light_buffer_offset = 0
            shadow_map_index = 0
            num_lights = 0
            light_buffers = viewport.scene_light_buffers.get_current_and_advance()
            for l in enabled_lights:
                if l.viewport_mask is None or l.viewport_mask & (1 << viewport_index):
                    if isinstance(l, DirectionalLight):
                        if num_lights >= self.max_lights_per_type:
                            raise RuntimeError(
                                f'Too many ligths of type: {LightTypes.DIRECTIONAL}. Increase "config.renderer.max_lights_per_type" (current value: {self.max_lights_per_type}) to allow more lights.'
                            )

                        map_index = -1
                        if l.shadow_map is not None:
                            map_index = shadow_map_index
                            shadow_map_index += 1

                            if map_index >= self.max_shadow_maps:
                                raise RuntimeError(
                                    f'Too many shadow_maps. Increase "config.renderer.max_shadow_maps" (current value: {self.max_shadow_maps}) to allow more shadow_maps.'
                                )

                        light_type_idx = LightTypes.DIRECTIONAL.value
                        offset = light_buffer_offset

                        l.upload_light(self, f, map_index, offset, light_buffers[light_type_idx])
                        light_buffer_offset += LIGHT_TYPES_INFO[light_type_idx].size
                        num_lights += 1
                    elif isinstance(l, EnvironmentLight):
                        if environment_light is not None:
                            raise RuntimeError(
                                f"More than one environment light enabled for viewport index {viewport_index}"
                            )
                        environment_light = l
                    elif isinstance(l, UniformEnvironmentLight):
                        uniform_environment_radiance += l.radiance.get_current()

            if path_tracing:
                assert viewport.path_tracer_viewport is not None
                assert self.path_tracer is not None
                buf = viewport.path_tracer_viewport.uniform_buffers.get_current_and_advance()
                constants = viewport.path_tracer_viewport.constants

                if viewport.camera.camera_from_world != viewport.path_tracer_viewport.last_camera_transform:
                    viewport.path_tracer_viewport.sample_index = 0
                    viewport.path_tracer_viewport.last_camera_transform = viewport.camera.camera_from_world.copy()

                camera_right, camera_up, camera_front = viewport.camera.right_up_front()
                constants["width"] = viewport_width
                constants["height"] = viewport_height
                constants["max_bounces"] = self.path_tracer_max_bounces
                constants["num_directional_lights"] = num_lights
                constants["camera_position"] = viewport.camera.position()
                constants["viewport_mask"] = 1 << viewport_index
                constants["camera_forward"] = camera_front
                constants["film_dist"] = viewport.camera.film_dist()
                constants["camera_up"] = camera_up
                constants["max_samples_per_pixel"] = self.path_tracer_max_samples_per_pixel
                constants["camera_right"] = camera_right
                constants["clip_value"] = self.path_tracer_clip_value
                constants["ambient_light"] = uniform_environment_radiance
                if environment_light is not None:
                    constants["has_environment_light"] = 1
                else:
                    constants["has_environment_light"] = 0
                constants["background_color"] = self.background_color[:3]
                constants["use_background_color"] = self.path_tracer_use_background_color
                buf.upload(
                    cmd,
                    MemoryUsage.NONE,
                    constants.view(np.uint8),
                )
            else:
                buf = viewport.scene_uniform_buffers.get_current_and_advance()
                proj = viewport.camera.projection(viewport_width / viewport_height)

                constants = viewport.scene_constants
                constants["projection"] = proj
                constants["view"] = viewport.camera.view()
                constants["world_camera_position"] = viewport.camera.position()
                constants["num_lights"] = num_lights
                constants["ambient_light"] = uniform_environment_radiance
                if environment_light is not None:
                    constants["has_environment_light"] = 1
                    constants["max_specular_mip"] = environment_light.gpu_cubemaps.specular_cubemap.mip_levels - 1.0
                else:
                    constants["has_environment_light"] = 0
                    constants["max_specular_mip"] = 0.0
                constants["inverse_viewport_size"] = (1.0 / viewport_width, 1.0 / viewport_height)
                constants["focal"] = (proj[0][0] * 0.5 * viewport_width, proj[1][1] * 0.5 * viewport_height)

                buf.upload(
                    cmd,
                    MemoryUsage.NONE,  # None because will be synced by after-upload barriers
                    constants.view(np.uint8),
                )

        self.enabled_gpu_properties = enabled_gpu_properties

        for o in enabled_objects:
            if o.material is not None:
                o.material.upload(self, f)
            o.upload(self, f)

        for w in enabled_widgets:
            w.upload(self, f)

        # Sync: after-upload transfer queue barriers (queue release)
        if frame.transfer_semaphores:
            assert frame.transfer_command_buffer is not None
            frame.transfer_command_buffer.barriers(
                buffer_barriers=f.transfer_upload_after_buffer_barriers,
                image_barriers=f.transfer_upload_after_image_barriers,
            )

        # Stage: in flight mip generation
        if f.mip_generation_requests:
            # Sync: after-upload image barriers before mip generation
            cmd.barriers(image_barriers=f.upload_after_image_barriers)

            # Batched mip creation.
            #
            # Assumes that every streaming mip object has a spd pipeline instance
            # preallocated, so that we can run all of them in the same submit.
            # This is the path that streaming and dynamic properties will take when
            # they need on the fly mip generation that does not stall.
            # Streaming properties will preallocate an spd pipeline instance while
            # preuploaded properties will allocate one when promoted to dynamic.
            mip_after_image_barriers = self.spd_pipeline.run_batched(f.cmd, f.mip_generation_requests)

            # Sync: after-upload barriers and post mip generation barriers
            cmd.barriers(
                memory_barriers=[
                    MemoryBarrier(
                        PipelineStageFlags.TRANSFER,
                        AccessFlags.TRANSFER_READ | AccessFlags.TRANSFER_WRITE,
                        f.upload_property_pipeline_stages,
                        AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                    ),
                ],
                buffer_barriers=f.upload_after_buffer_barriers,
                image_barriers=mip_after_image_barriers,
            )
        else:
            # Sync: after-upload barriers
            cmd.barriers(
                memory_barriers=[
                    MemoryBarrier(
                        PipelineStageFlags.TRANSFER,
                        AccessFlags.TRANSFER_READ | AccessFlags.TRANSFER_WRITE,
                        f.upload_property_pipeline_stages,
                        AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                    ),
                ],
                buffer_barriers=f.upload_after_buffer_barriers,
                image_barriers=f.upload_after_image_barriers,
            )

        # Stage: post upload
        for o in enabled_objects:
            o.post_upload(self, f)

        # Create and upload path tracing acceleration structures
        if path_tracing:
            assert self.path_tracer is not None

            # Create or resize accumulation buffer
            for v in viewports:
                assert v.path_tracer_viewport is not None
                if (
                    v.path_tracer_viewport.accumulation_image is None
                    or self.render_width != v.path_tracer_viewport.accumulation_image.width
                    or self.render_height != v.path_tracer_viewport.accumulation_image.height
                ):
                    # Assume no rendering is in flight if resolution changed
                    if v.path_tracer_viewport.accumulation_image is not None:
                        v.path_tracer_viewport.accumulation_image.destroy()

                    v.path_tracer_viewport.accumulation_image = Image(
                        self.device,
                        self.render_width,
                        self.render_height,
                        Format.R32G32B32A32_SFLOAT,
                        ImageUsageFlags.STORAGE,
                        AllocType.DEVICE_DEDICATED,
                        name="accumulation",
                    )
                    cmd.image_barrier(
                        v.path_tracer_viewport.accumulation_image,
                        ImageLayout.GENERAL,
                        MemoryUsage.NONE,
                        MemoryUsage.COMPUTE_SHADER,
                    )
                    for s in v.path_tracer_viewport.descriptor_sets:
                        s.write_image(
                            v.path_tracer_viewport.accumulation_image,
                            ImageLayout.GENERAL,
                            DescriptorType.STORAGE_IMAGE,
                            4,
                        )

            # Stage: path tracing structures creation
            if self.path_tracer.acceleration_structure is None:
                # Flush commands and re-open per-frame command buffers.
                #
                # We need this because acceleration structure creation could use just uploaded positions and indices.
                # Since we also want to render a frame this frame we ensure that all uploads happened and then reset
                # command buffers.
                #
                # We could skip this if we knew that we don't have any positions or indices uploaded this frame
                # but there is no simple way to detect that right now.
                # A better system would have each primitive manage the creation and update of their acceleration
                # structures and avoid this kind of stall.
                # Since we are now anyways stalling the pipeline to create acceleration structures this overhead
                # should be negligible compared to the actual build time and synchronization.
                f.cmd.end()
                self.device.graphics_queue.submit(
                    f.cmd,
                    wait_semaphores=[(s.sem, s.wait_value, s.wait_stage) for s in f.additional_semaphores],
                    signal_semaphores=[(s.sem, s.signal_value, s.signal_stage) for s in f.additional_semaphores],
                )
                if f.transfer_semaphores:
                    assert f.transfer_cmd is not None
                    f.transfer_cmd.end()
                    self.device.transfer_queue.submit(
                        f.transfer_cmd,
                        wait_semaphores=[(s.sem, s.wait_value, s.wait_stage) for s in f.transfer_semaphores],
                        signal_semaphores=[(s.sem, s.signal_value, s.signal_stage) for s in f.transfer_semaphores],
                    )

                # Wait for commands to complete
                self.device.wait_idle()

                # Reset commands and semaphores
                f.cmd.begin()
                if f.transfer_semaphores:
                    assert f.transfer_cmd is not None
                    f.transfer_cmd.begin()
                f.additional_semaphores.clear()
                f.transfer_semaphores.clear()

                instances: List[AccelerationStructureInstanceInfo] = []
                materials: Dict[Material, int] = {}
                for o in enabled_objects:
                    if o.material is not None:
                        if o.material in materials:
                            material_index = materials[o.material]
                        else:
                            material_index = len(materials)
                            materials[o.material] = material_index
                        o.append_acceleration_structure_instances(instances, material_index)

                if not instances:
                    if self.zero_triangle is None:
                        self.zero_triangle = Buffer.from_data(
                            self.device,
                            np.array([0.0, 0.0, 0.0], np.float32),
                            BufferUsageFlags.SHADER_DEVICE_ADDRESS
                            | BufferUsageFlags.ACCELERATION_STRUCTURE_INPUT
                            | BufferUsageFlags.TRANSFER_DST,
                            AllocType.DEVICE_MAPPED_WITH_FALLBACK,
                            "zero-triangle",
                        )
                    instances.append(
                        AccelerationStructureInstanceInfo(
                            np.zeros((3, 4), np.float32),
                            np.zeros((3, 4), np.float32),
                            3,
                            self.zero_triangle.address,
                            0,
                            0,
                            0,
                            1,
                            0,
                            0,
                            0,
                        )
                    )
                if not materials:
                    if self.zero_material is None:
                        self.zero_material = ColorMaterial((0, 0, 0))
                    materials[self.zero_material] = 0

                # Instances and acceleration structures
                acceleration_structure_meshes: List[AccelerationStructureMesh] = []
                instances_data = np.zeros(len(instances), self.path_tracer_instance_dtype)
                for i, instance in enumerate(instances):
                    instances_data[i]["normal_matrix"] = instance.normal_matrix
                    instances_data[i]["positions"] = instance.positions_address
                    instances_data[i]["normals"] = instance.normals_address
                    instances_data[i]["tangents"] = instance.tangents_address
                    instances_data[i]["uvs"] = instance.uvs_address
                    instances_data[i]["indices"] = instance.indices_address
                    instances_data[i]["material_index"] = instance.material_index

                    acceleration_structure_meshes.append(
                        AccelerationStructureMesh(
                            instance.positions_address,
                            12,
                            instance.positions_count,
                            Format.R32G32B32_SFLOAT,
                            instance.indices_address,
                            IndexType.UINT32 if instance.indices_address else IndexType.NONE,
                            instance.primitive_count,
                            tuple(instance.transform.flatten()),
                            instance.viewport_mask,
                        )
                    )
                self.path_tracer.acceleration_structure = AccelerationStructure(
                    self.device, acceleration_structure_meshes, name="acceleration-structure"
                )
                self.path_tracer.instances_buf = Buffer.from_data(
                    self.device,
                    view_bytes(instances_data),
                    BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
                    AllocType.DEVICE_MAPPED_WITH_FALLBACK,
                    name="path-tracer-instances",
                )

                # Upload materials and write bindless descriptors
                textures: Dict[Union[Image, ImageView], int] = {}
                materials_data = np.zeros(len(materials), self.path_tracer_material_dtype)

                def alloc_texture(p: Property) -> int:
                    if isinstance(p, ImageProperty):
                        view = p.get_current_gpu().view()
                        if view in textures:
                            texture_index = textures[view]
                        else:
                            texture_index = len(textures)
                            textures[view] = texture_index
                        return texture_index
                    else:
                        return 0xFFFFFFFF

                def get_value(p: Property) -> Union[float, PropertyItem]:
                    if isinstance(p, BufferProperty):
                        return p.get_current()
                    else:
                        return 0.0

                for i, material in enumerate(materials.keys()):
                    if isinstance(material, DiffuseMaterial):
                        materials_data[i]["albedo"] = get_value(material.diffuse[0].property)
                        materials_data[i]["albedo_texture"] = alloc_texture(material.diffuse[0].property)
                        materials_data[i]["roughness"] = 1.0
                        materials_data[i]["roughness_texture"] = 0xFFFFFFFF
                        materials_data[i]["metallic"] = 0.0
                        materials_data[i]["metallic_texture"] = 0xFFFFFFFF
                        materials_data[i]["normal_texture"] = alloc_texture(material.normal[0].property)
                    elif isinstance(material, DiffuseSpecularMaterial):
                        # TODO: better mapping or remove
                        materials_data[i]["albedo"] = get_value(material.diffuse[0].property)
                        materials_data[i]["albedo_texture"] = alloc_texture(material.diffuse[0].property)
                        materials_data[i]["roughness"] = (
                            1 - min(get_value(material.specular_exponent[0].property), 64.0) / 64.0
                        )
                        materials_data[i]["roughness_texture"] = 0xFFFFFFFF
                        materials_data[i]["metallic"] = 0.0
                        materials_data[i]["metallic_texture"] = 0xFFFFFFFF
                        materials_data[i]["normal_texture"] = 0xFFFFFFFF
                    elif isinstance(material, ColorMaterial):
                        # TODO: use emissive instead
                        materials_data[i]["albedo"] = get_value(material.color[0].property)
                        materials_data[i]["albedo_texture"] = alloc_texture(material.color[0].property)
                        materials_data[i]["roughness"] = 1.0
                        materials_data[i]["roughness_texture"] = 0xFFFFFFFF
                        materials_data[i]["metallic"] = 0.0
                        materials_data[i]["metallic_texture"] = 0xFFFFFFFF
                        materials_data[i]["normal_texture"] = 0xFFFFFFFF
                    elif isinstance(material, PBRMaterial):
                        materials_data[i]["albedo"] = get_value(material.albedo[0].property)
                        materials_data[i]["albedo_texture"] = alloc_texture(material.albedo[0].property)
                        materials_data[i]["roughness"] = get_value(material.roughness[0].property)
                        materials_data[i]["roughness_texture"] = alloc_texture(material.roughness[0].property)
                        materials_data[i]["metallic"] = get_value(material.metallic[0].property)
                        materials_data[i]["metallic_texture"] = alloc_texture(material.metallic[0].property)
                        materials_data[i]["normal_texture"] = alloc_texture(material.normal[0].property)
                    else:
                        materials_data[i]["albedo"] = 0.0
                        materials_data[i]["albedo_texture"] = 0xFFFFFFFF
                        materials_data[i]["roughness"] = 1.0
                        materials_data[i]["roughness_texture"] = 0xFFFFFFFF
                        materials_data[i]["metallic"] = 0.0
                        materials_data[i]["metallic_texture"] = 0xFFFFFFFF
                        materials_data[i]["normal_texture"] = 0xFFFFFFFF
                if self.path_tracer.materials_buf is not None:
                    self.path_tracer.materials_buf.destroy()
                self.path_tracer.materials_buf = Buffer.from_data(
                    self.device,
                    view_bytes(materials_data),
                    BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
                    AllocType.DEVICE_MAPPED_WITH_FALLBACK,
                    name="path-tracer-materials",
                )

                # Static descriptors
                self.path_tracer.descriptor_set.write_acceleration_structure(
                    self.path_tracer.acceleration_structure, 0
                )
                self.path_tracer.descriptor_set.write_buffer(
                    self.path_tracer.instances_buf, DescriptorType.STORAGE_BUFFER, 1
                )
                self.path_tracer.descriptor_set.write_buffer(
                    self.path_tracer.materials_buf, DescriptorType.STORAGE_BUFFER, 2
                )
                self.path_tracer.descriptor_set.write_sampler(self.linear_sampler, 3)
                for i, image in enumerate(textures):
                    self.path_tracer.descriptor_set.write_image(
                        image, ImageLayout.SHADER_READ_ONLY_OPTIMAL, DescriptorType.SAMPLED_IMAGE, 4, i
                    )

        # Stage: Render shadows
        if not path_tracing:
            for l in enabled_lights:
                l.render_shadow_maps(self, f, enabled_objects)

        before_gui_barriers: List[ImageBarrier] = []
        if path_tracing:
            if self.multiviewport:
                f.before_render_image_barriers.append(
                    ImageBarrier(
                        frame.image,
                        ImageLayout.UNDEFINED,
                        ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                    )
                )
                for v in viewports:
                    assert v.image is not None
                    f.before_render_image_barriers.append(
                        ImageBarrier(
                            v.image,
                            ImageLayout.UNDEFINED,
                            ImageLayout.GENERAL,
                            PipelineStageFlags.FRAGMENT_SHADER,
                            AccessFlags.SHADER_SAMPLED_READ,
                            PipelineStageFlags.COMPUTE_SHADER,
                            AccessFlags.SHADER_STORAGE_READ | AccessFlags.SHADER_STORAGE_WRITE,
                        )
                    )
                    before_gui_barriers.append(
                        ImageBarrier(
                            v.image,
                            ImageLayout.GENERAL,
                            ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                            PipelineStageFlags.COMPUTE_SHADER,
                            AccessFlags.SHADER_STORAGE_READ | AccessFlags.SHADER_STORAGE_WRITE,
                            PipelineStageFlags.FRAGMENT_SHADER,
                            AccessFlags.SHADER_SAMPLED_READ,
                        )
                    )
            else:
                f.before_render_image_barriers.append(
                    ImageBarrier(
                        frame.image,
                        ImageLayout.UNDEFINED,
                        ImageLayout.GENERAL,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                        PipelineStageFlags.COMPUTE_SHADER,
                        AccessFlags.SHADER_STORAGE_READ | AccessFlags.SHADER_STORAGE_WRITE,
                    )
                )
                before_gui_barriers.append(
                    ImageBarrier(
                        frame.image,
                        ImageLayout.GENERAL,
                        ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                        PipelineStageFlags.COMPUTE_SHADER,
                        AccessFlags.SHADER_STORAGE_READ | AccessFlags.SHADER_STORAGE_WRITE,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                    )
                )
        else:
            f.before_render_image_barriers.extend(
                [
                    ImageBarrier(
                        frame.image,
                        ImageLayout.UNDEFINED,
                        ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                    ),
                    ImageBarrier(
                        self.depth_buffer,
                        ImageLayout.UNDEFINED,
                        ImageLayout.DEPTH_STENCIL_ATTACHMENT_OPTIMAL,
                        PipelineStageFlags.LATE_FRAGMENT_TESTS,
                        AccessFlags.DEPTH_STENCIL_ATTACHMENT_READ | AccessFlags.DEPTH_STENCIL_ATTACHMENT_WRITE,
                        PipelineStageFlags.EARLY_FRAGMENT_TESTS,
                        AccessFlags.DEPTH_STENCIL_ATTACHMENT_READ | AccessFlags.DEPTH_STENCIL_ATTACHMENT_WRITE,
                        aspect_mask=ImageAspectFlags.DEPTH,
                    ),
                ]
            )

            if self.multiviewport:
                for v in viewports:
                    assert v.image is not None
                    f.before_render_image_barriers.append(
                        ImageBarrier(
                            v.image,
                            ImageLayout.UNDEFINED,
                            ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                            PipelineStageFlags.FRAGMENT_SHADER,
                            AccessFlags.SHADER_SAMPLED_READ,
                            PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                            AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                        )
                    )
                    before_gui_barriers.append(
                        ImageBarrier(
                            v.image,
                            ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                            ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                            PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                            AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                            PipelineStageFlags.FRAGMENT_SHADER,
                            AccessFlags.SHADER_SAMPLED_READ | AccessFlags.SHADER_SAMPLED_READ,
                        )
                    )

            if self.msaa_target is not None:
                f.before_render_image_barriers.append(
                    ImageBarrier(
                        self.msaa_target,
                        ImageLayout.UNDEFINED,
                        ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                        PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                        AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                    )
                )

        # Sync: before-render barriers
        cmd.barriers(
            memory_barriers=[
                MemoryBarrier(
                    f.before_render_src_pipeline_stages,
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                    f.before_render_dst_pipeline_stages,
                    AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                ),
            ],
            image_barriers=f.before_render_image_barriers,
        )

        for viewport_index, viewport in enumerate(viewports):
            if viewport.rect.width <= 0 or viewport.rect.height <= 0:
                continue
            viewport_bit = 1 << viewport_index

            if path_tracing:
                assert self.path_tracer is not None
                assert viewport.path_tracer_viewport is not None
                descriptor_set = viewport.path_tracer_viewport.descriptor_sets.get_current_and_advance()
                written_environment = False
                for l in enabled_lights:
                    if l.viewport_mask is None or l.viewport_mask & (1 << viewport_index):  # noqa: SIM102
                        if isinstance(l, EnvironmentLight):
                            descriptor_set.write_image(
                                l.gpu_cubemaps.specular_cubemap,
                                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                                DescriptorType.SAMPLED_IMAGE,
                                3,
                            )
                            written_environment = True
                if not written_environment:
                    descriptor_set.write_image(
                        self.zero_cubemap,
                        ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                        DescriptorType.SAMPLED_IMAGE,
                        3,
                    )
                descriptor_set.write_image(
                    frame.image if viewport.image is None else viewport.image,
                    ImageLayout.GENERAL,
                    DescriptorType.STORAGE_IMAGE,
                    2,
                )
                cmd.bind_compute_pipeline(self.path_tracer.pipeline, [self.path_tracer.descriptor_set, descriptor_set])
                for _ in range(self.path_tracer_samples_per_frame):
                    self.path_tracer_push_constants["sample_index"] = viewport.path_tracer_viewport.sample_index
                    cmd.push_constants(self.path_tracer.pipeline, self.path_tracer_push_constants.tobytes())
                    cmd.dispatch(div_round_up(viewport.rect.width, 8), div_round_up(viewport.rect.height, 8), 1)
                    cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.COMPUTE_SHADER)
                    if viewport.path_tracer_viewport.sample_index < self.path_tracer_max_samples_per_pixel:
                        viewport.path_tracer_viewport.sample_index += 1
                    if viewport.path_tracer_viewport.sample_index >= self.path_tracer_max_samples_per_pixel:
                        break
            else:
                descriptor_set = viewport.scene_descriptor_sets.get_current_and_advance()
                shadow_map_index = 0
                for l in enabled_lights:
                    if l.viewport_mask is None or l.viewport_mask & (1 << viewport_index):
                        if isinstance(l, DirectionalLight) and l.shadow_map is not None:
                            descriptor_set.write_image(
                                l.shadow_map,
                                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                                DescriptorType.SAMPLED_IMAGE,
                                6 + len(LIGHT_TYPES_INFO),
                                shadow_map_index,
                            )
                            shadow_map_index += 1
                        elif isinstance(l, EnvironmentLight):
                            descriptor_set.write_image(
                                l.gpu_cubemaps.irradiance_cubemap,
                                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                                DescriptorType.SAMPLED_IMAGE,
                                3,
                            )
                            descriptor_set.write_image(
                                l.gpu_cubemaps.specular_cubemap,
                                ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                                DescriptorType.SAMPLED_IMAGE,
                                4,
                            )

                if viewport_index > 0 and f.between_viewport_render_src_pipeline_stages:
                    cmd.barriers(
                        memory_barriers=[
                            MemoryBarrier(
                                f.between_viewport_render_src_pipeline_stages,
                                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                                f.between_viewport_render_dst_pipeline_stages,
                                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                            ),
                        ],
                    )

                # Stage: pre-render
                for o in enabled_objects:
                    if o.viewport_mask is None or (o.viewport_mask & viewport_bit):
                        o.pre_render(self, f, viewport, descriptor_set)

                # Sync: before-render barriers
                if f.before_render_src_pipeline_stages:
                    cmd.barriers(
                        memory_barriers=[
                            MemoryBarrier(
                                f.before_render_src_pipeline_stages,
                                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                                f.before_render_dst_pipeline_stages,
                                AccessFlags.MEMORY_READ | AccessFlags.MEMORY_WRITE,
                            ),
                        ],
                    )

                viewport_rect = (
                    0,
                    viewport.rect.height,
                    viewport.rect.width,
                    -viewport.rect.height,
                )
                rect = (
                    0,
                    0,
                    viewport.rect.width,
                    viewport.rect.height,
                )

                # Stage: render
                cmd.set_viewport(viewport_rect)
                cmd.set_scissors(rect)

                image = frame.image if viewport.image is None else viewport.image
                with cmd.rendering(
                    rect,
                    color_attachments=[
                        (
                            RenderingAttachment(
                                self.msaa_target,
                                load_op=LoadOp.CLEAR,
                                store_op=StoreOp.STORE,
                                clear=self.background_color,
                                resolve_mode=ResolveMode.AVERAGE,
                                resolve_image=image,
                            )
                            if self.msaa_target is not None
                            else RenderingAttachment(image, LoadOp.CLEAR, StoreOp.STORE, self.background_color)
                        )
                    ],
                    depth=DepthAttachment(self.depth_buffer, LoadOp.CLEAR, StoreOp.STORE, self.depth_clear_value),
                ):
                    for o in enabled_objects:
                        if o.viewport_mask is None or (o.viewport_mask & viewport_bit):
                            o.render(self, f, viewport, descriptor_set)

        if before_gui_barriers:
            cmd.barriers(image_barriers=before_gui_barriers)

        self.uniform_pool.advance()
        self.total_frame_index += 1

    def render_gui(self, frame: FrameInputs, gui: Gui) -> None:
        cmd = frame.command_buffer

        # Stage: Render GUI
        with cmd.rendering(
            (0, 0, frame.image.width, frame.image.height),
            color_attachments=[
                RenderingAttachment(
                    frame.image,
                    load_op=LoadOp.CLEAR if self.multiviewport else LoadOp.LOAD,
                    store_op=StoreOp.STORE,
                    clear=self.background_color,
                ),
            ],
        ):
            gui.render(cmd)

        # Sync: after render barriers
        cmd.image_barrier_full(
            ImageBarrier(
                frame.image,
                ImageLayout.COLOR_ATTACHMENT_OPTIMAL,
                ImageLayout.PRESENT_SRC,
                PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                AccessFlags.COLOR_ATTACHMENT_WRITE | AccessFlags.COLOR_ATTACHMENT_READ,
                PipelineStageFlags.COLOR_ATTACHMENT_OUTPUT,
                AccessFlags.NONE,
            )
        )

    def prefetch(self) -> None:
        for p in self.enabled_gpu_properties:
            p.prefetch()

    def enqueue_for_destruction(self, resources: Iterable[Union[Buffer, Image, ImageView]]) -> None:
        self.destruction_queue.setdefault(self.total_frame_index + self.num_frames_in_flight, []).extend(resources)

    def wait_and_destroy(self) -> None:
        self.device.wait_idle()
        for resources in self.destruction_queue.values():
            for resource in resources:
                resource.destroy()
        self.destruction_queue.clear()

    def toggle_path_tracer(self) -> None:
        if self.render_mode == RenderMode.RASTER:
            if self.supports_path_tracing:
                self.render_mode = RenderMode.PATH_TRACER
            else:
                logger.warning(
                    "Path tracing cannot be enabled because DeviceFeatures.RAY_QUERY is not supported. Falling back to raster mode."
                )
        else:
            self.render_mode = RenderMode.RASTER



import numpy as np

from ambra.config import Config, GuiConfig, PlaybackConfig, RendererConfig
from ambra.geometry import create_axis3d_lines_and_colors
from ambra.primitives3d import Lines
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        world_up=(0, 1, 0),
        gui=GuiConfig(
            multiviewport=True,
            initial_number_of_viewports=2,
            stats=True,
            inspector=True,
        ),
        renderer=RendererConfig(
            msaa_samples=4,
        ),
    ),
)

line = Lines(*create_axis3d_lines_and_colors(), 4)

viewer.scene.objects.extend([line])

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from typing import Optional, Union

import numpy as np
from pyglm.glm import mat3
from pyxpg import (
    Attachment,
    BufferUsageFlags,
    Depth,
    DescriptorSet,
    DescriptorSetBinding,
    DescriptorType,
    DeviceFeatures,
    Filter,
    Format,
    GraphicsPipeline,
    ImageLayout,
    ImageUsageFlags,
    InputAssembly,
    PipelineStage,
    PipelineStageFlags,
    PrimitiveTopology,
    PushConstantsRange,
    Rasterization,
    Sampler,
    SamplerAddressMode,
    Shader,
    Stage,
    VertexAttribute,
    VertexBinding,
    VertexInputRate,
)

from .property import BufferProperty, ImageProperty
from .renderer import Renderer
from .renderer_frame import RendererFrame
from .scene import Object, Object2D
from .utils.descriptors import create_descriptor_layout_pool_and_sets_ringbuffer
from .viewport import Viewport


class Lines(Object2D):
    def __init__(
        self,
        lines: BufferProperty,
        colors: BufferProperty,
        line_width: Union[BufferProperty, float] = 1.0,
        is_strip: bool = False,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)

        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.is_strip = is_strip
        self.lines = self.add_buffer_property(lines, np.float32, (-1, 2), name="lines").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.colors = self.add_buffer_property(colors, np.uint32, (-1,), name="colors").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.line_width = self.add_buffer_property(line_width, np.float32, name="line_width")

    def create(self, r: Renderer) -> None:
        vert = r.compile_builtin_shader("2d/basic.slang", "vertex_main")
        frag = r.compile_builtin_shader("2d/basic.slang", "pixel_main")

        # Instantiate the pipeline using the compiled shaders
        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=[
                VertexBinding(0, 8, VertexInputRate.VERTEX),
                VertexBinding(1, 4, VertexInputRate.VERTEX),
            ],
            vertex_attributes=[
                VertexAttribute(0, 0, Format.R32G32_SFLOAT),
                VertexAttribute(1, 1, Format.R32_UINT),
            ],
            rasterization=Rasterization(dynamic_line_width=True),
            input_assembly=InputAssembly(
                PrimitiveTopology.LINE_STRIP if self.is_strip else PrimitiveTopology.LINE_LIST
            ),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, False, False, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"][:, :3, :3] = mat3(self.current_transform_matrix)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        lines = self.lines.get_current_gpu()
        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=[
                lines.buffer_and_offset(),
                self.colors.get_current_gpu().buffer_and_offset(),
            ],
            descriptor_sets=[
                scene_descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )
        frame.cmd.set_line_width(
            self.line_width.get_current().item() if renderer.device.features & DeviceFeatures.WIDE_LINES else 1.0
        )

        frame.cmd.draw(lines.size // 8)


class Image(Object2D):
    def __init__(
        self,
        image: ImageProperty,
        mips: bool = False,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)
        super().__init__(name, translation, rotation, scale, enabled=enabled, viewport_mask=viewport_mask)
        self.image = self.add_image_property(image, name="image").use_gpu(
            ImageUsageFlags.SAMPLED,
            ImageLayout.SHADER_READ_ONLY_OPTIMAL,
            PipelineStageFlags.FRAGMENT_SHADER,
            mips=mips,
        )

    def create(self, r: Renderer) -> None:
        if not (r.device.features & DeviceFeatures.SHADER_DRAW_PARAMETERS):
            raise RuntimeError(
                f"Image primitive requires {DeviceFeatures.SHADER_DRAW_PARAMETERS} whis is not available on current device."
            )

        self.sampler = Sampler(
            r.device,
            min_filter=Filter.LINEAR,
            mag_filter=Filter.LINEAR,
            u=SamplerAddressMode.CLAMP_TO_EDGE,
            v=SamplerAddressMode.CLAMP_TO_EDGE,
        )

        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_sets = (
            create_descriptor_layout_pool_and_sets_ringbuffer(
                r.device,
                [
                    DescriptorSetBinding(1, DescriptorType.SAMPLER),
                    DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE),
                ],
                r.num_frames_in_flight,
                name=f"{self.name}-descriptors",
            )
        )
        for set in self.descriptor_sets:
            set.write_sampler(self.sampler, 0)

        vert = r.compile_builtin_shader("2d/basic_texture.slang", "vertex_main")
        frag = r.compile_builtin_shader("2d/basic_texture.slang", "pixel_main")

        # Instantiate the pipeline using the compiled shaders
        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            input_assembly=InputAssembly(PrimitiveTopology.TRIANGLE_STRIP),
            samples=r.msaa_samples,
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, False, False, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def update_transform(self, parent: Optional[Object]) -> None:
        super().update_transform(parent)
        self.constants["transform"][:, :3, :3] = mat3(self.current_transform_matrix)

    def render(
        self, renderer: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        descriptor_set = self.descriptor_sets.get_current_and_advance()
        descriptor_set.write_image(
            self.image.get_current_gpu().view(),
            ImageLayout.SHADER_READ_ONLY_OPTIMAL,
            DescriptorType.SAMPLED_IMAGE,
            1,
        )

        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            descriptor_sets=[
                scene_descriptor_set,
                descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )

        frame.cmd.draw(4)



import numpy as np
from pyglm.glm import vec3

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig
from ambra.geometry import create_axis3d_lines_and_colors
from ambra.primitives3d import Lines
from ambra.property import (
    AnimationBoundary,
    ArrayBufferProperty,
    FrameAnimation,
)
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        camera=CameraConfig(
            position=vec3(3),
            target=vec3(0),
        ),
        gui=GuiConfig(
            playback=True,
            stats=True,
            inspector=True,
        ),
    ),
)

positions, colors = create_axis3d_lines_and_colors()
positions = np.linspace(positions, positions + np.array([0, 0, 0]), 100)
line_width = np.linspace(1, 32, 100)

translation = ArrayBufferProperty(
    np.linspace(np.array([0, 0, 0]), np.array([0, 1, 1]), 50),
    np.float32,
    FrameAnimation(boundary=AnimationBoundary.MIRROR),
)

line = Lines(positions, colors, line_width, scale=None, translation=translation)

translation = ArrayBufferProperty(
    np.linspace(np.array([1, 0, 0]), np.array([2, 0, 0]), 100),
    np.float32,
    FrameAnimation(start_frame=10, boundary=AnimationBoundary.MIRROR),
)
line2 = Lines(positions, colors, line_width, scale=None, translation=translation)
viewer.scene.objects.extend([line, line2])

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass

from pyglm.glm import mat3, rotate, vec2


@dataclass
class RigidTransform2D:
    translation: vec2
    rotation: float

    @classmethod
    def identity(cls) -> "RigidTransform2D":
        return cls(
            translation=vec2(0),
            rotation=0,
        )


@dataclass
class Transform2D:
    translation: vec2
    rotation: float
    scale: vec2

    @classmethod
    def identity(cls) -> "Transform2D":
        return cls(
            translation=vec2(0),
            rotation=0,
            scale=vec2(1),
        )

    def as_mat3(self) -> mat3:
        m = rotate(self.rotation)
        m[0] *= self.scale[0]
        m[1] *= self.scale[1]
        m[2, 0] = self.translation.x
        m[2, 1] = self.translation.y
        return m



import numpy as np

from ambra.config import Config, GuiConfig
from ambra.viewer import Viewer
from ambra.widgets import ImageInspector

viewer = Viewer(
    config=Config(
        gui=GuiConfig(
            stats=True,
        ),
    ),
)


H, W = 8, 16
img_data = np.zeros((H, W, 4))
for i in range(H):
    for j in range(W):
        img_data[i, j, 0] = (i + 0.5) / H
        img_data[i, j, 1] = (j + 0.5) / W
        img_data[i, j, 3] = 1
img_data = (img_data * 255.0).astype(np.uint8)

img_inspector = ImageInspector("Image Inspector", img_data)

viewer.scene.widgets.append(img_inspector)

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import math
from dataclasses import dataclass
from typing import Any, List, Optional

from numpy.typing import NDArray
from pyglm.glm import (
    acos,
    clamp,
    cross,
    distance,
    dot,
    dvec2,
    ivec2,
    mat3,
    mat3_cast,
    mat4,
    normalize,
    rotate,
    sqrt,
    translate,
    transpose,
    vec2,
    vec3,
    vec4,
)
from pyxpg import DescriptorSet, Image, imgui

from .camera import Camera, CameraDepth, OrthographicCamera, PerspectiveCamera
from .config import CameraConfig, CameraControlMode, CameraProjection, Handedness, PlaybackConfig
from .transform3d import RigidTransform3D
from .utils.gpu import UploadableBuffer
from .utils.ring_buffer import RingBuffer


class Playback:
    def __init__(self, config: PlaybackConfig):
        self.max_time = config.max_time
        self.num_frames = int(self.max_time * config.frames_per_second) if self.max_time is not None else 0

        self.playing = config.playing
        self.frames_per_second = config.frames_per_second
        self.playback_speed_multiplier = config.playback_speed_multiplier

        if config.initial_frame is None:
            self.current_time = config.initial_time
            self.current_frame = int(self.current_time * self.frames_per_second)
        else:
            self.set_frame(config.initial_frame)

    def set_max_time(self, max_time: float) -> None:
        self.max_time = max(max_time, 0.0)
        self.num_frames = max(int(self.max_time * self.frames_per_second), 1)

    def step(self, dt: float) -> None:
        self.set_time(self.current_time + dt * self.playback_speed_multiplier)

    def toggle_play_pause(self) -> None:
        self.playing = not self.playing

    def set_time(self, time: float) -> None:
        assert self.max_time is not None
        time = max(time, 0.0)
        self.current_time = math.fmod(time, self.max_time) if self.max_time != 0.0 else 0.0
        self.current_frame = int(self.current_time * self.frames_per_second)

    def set_frame(self, frame: int) -> None:
        self.current_frame = frame % self.num_frames
        self.current_time = self.current_frame / self.frames_per_second


@dataclass
class Rect:
    x: int
    y: int
    width: int
    height: int


@dataclass
class PathTracerViewport:
    descriptor_sets: RingBuffer[DescriptorSet]
    uniform_buffers: RingBuffer[UploadableBuffer]
    constants: NDArray[Any]
    sample_index: int
    last_camera_transform: RigidTransform3D
    accumulation_image: Optional[Image]


class Viewport:
    def __init__(
        self,
        rect: Rect,
        playback: Playback,
        camera_config: CameraConfig,
        world_up: vec3,
        handedness: Handedness,
        scene_descriptor_sets: RingBuffer[DescriptorSet],
        scene_uniform_buffers: RingBuffer[UploadableBuffer],
        scene_light_buffers: RingBuffer[List[UploadableBuffer]],
        scene_constants: NDArray[Any],
        image: Optional[Image],
        imgui_texture: Optional[imgui.Texture],
        name: str,
    ):
        camera_from_world = RigidTransform3D.look_at(
            vec3(camera_config.position),
            vec3(camera_config.target),
            world_up,
            handedness,
        )

        camera: Camera
        if camera_config.projection == CameraProjection.PERSPECTIVE:
            camera = PerspectiveCamera(
                camera_from_world,
                CameraDepth(camera_config.z_near, camera_config.z_far),
                camera_config.perspective_vertical_fov,
            )
        elif camera_config.projection == CameraProjection.ORTHOGRAPHIC:
            camera = OrthographicCamera(
                camera_from_world,
                CameraDepth(camera_config.z_near, camera_config.z_far),
                vec2(camera_config.ortho_center),
                vec2(camera_config.ortho_half_extents),
            )
        else:
            raise RuntimeError(f"Unhandled camera type {camera_config.projection}")

        self.rect = rect
        self.camera = camera
        self.playback = playback
        self.scene_descriptor_sets = scene_descriptor_sets
        self.scene_uniform_buffers = scene_uniform_buffers
        self.scene_light_buffers = scene_light_buffers
        self.scene_constants = scene_constants
        self.image = image
        self.imgui_texture = imgui_texture
        self.name = name

        # Path tracer
        self.path_tracer_viewport: Optional[PathTracerViewport] = None

        # Config
        self.handedness = handedness
        self.camera_world_up = world_up
        self.camera_target = vec3(camera_config.target)
        self.camera_control_mode = camera_config.control_mode
        self.camera_rotation_speed = vec2(camera_config.rotation_speed)
        self.camera_pan_speed = vec2(camera_config.pan_speed)
        self.camera_pan_distance_speed_scale = camera_config.pan_distance_speed_scale
        self.camera_pan_min_speed_scale = camera_config.pan_min_speed_scale
        self.camera_zoom_speed = camera_config.zoom_speed
        self.camera_zoom_distance_speed_scale = camera_config.zoom_distance_speed_scale
        self.camera_zoom_min_speed_scale = camera_config.zoom_min_speed_scale
        self.camera_zoom_min_target_distance = camera_config.zoom_min_target_distance
        self.camera_zoom_ortho_speed = camera_config.zoom_ortho_speed

        # State
        self.rotate_pressed = False
        self.pan_pressed = False
        self.drag_start_mouse_position = ivec2(0, 0)
        self.drag_start_camera_inverse_view_rotation = mat3()
        self.drag_start_camera_position = vec3(0)
        self.drag_start_camera_target = vec3(0)
        self.drag_start_camera_right = vec3(0)
        self.drag_start_camera_up = vec3(0)
        self.drag_start_camera_pitch = 0.0

    def resize(self, width: int, height: int) -> None:
        if self.path_tracer_viewport is not None and (self.rect.width != width or self.rect.height != height):
            # Invalidate accumulation after resize
            self.path_tracer_viewport.sample_index = 0
        self.rect.width = width
        self.rect.height = height

    def start_drag(self, position: ivec2) -> None:
        self.drag_start_mouse_position = position
        self.drag_start_camera_inverse_view_rotation = transpose(mat3_cast(self.camera.camera_from_world.rotation))  # type: ignore
        self.drag_start_camera_position = self.camera.position()
        self.drag_start_camera_target = self.camera_target
        right, up, front = self.camera.right_up_front()
        self.drag_start_camera_right = right
        self.drag_start_camera_up = up
        d = -dot(front, self.camera_world_up)
        self.drag_start_camera_pitch = acos(clamp(d, -1, 1))

    def on_rotate_press(self, position: ivec2) -> None:
        self.rotate_pressed = True
        self.pan_pressed = False
        self.start_drag(position)

    def on_rotate_release(self) -> None:
        self.rotate_pressed = False

    def on_pan_press(self, position: ivec2) -> None:
        self.pan_pressed = True
        self.rotate_pressed = False
        self.start_drag(position)

    def on_pan_release(self) -> None:
        self.pan_pressed = False

    def on_move(self, position: ivec2) -> None:
        if self.pan_pressed:
            self.pan(self.drag_start_mouse_position, position)
        if self.rotate_pressed:
            if self.camera_control_mode == CameraControlMode.ORBIT:
                self.rotate_orbit(self.drag_start_mouse_position, position)
            elif self.camera_control_mode == CameraControlMode.TRACKBALL:
                self.rotate_trackball(self.drag_start_mouse_position, position)
            elif self.camera_control_mode == CameraControlMode.FIRST_PERSON:
                self.rotate_first_person(self.drag_start_mouse_position, position)
            elif (
                # self.camera_control_mode == CameraControlMode.PAN_AND_ZOOM_ORTHO or
                self.camera_control_mode == CameraControlMode.NONE
            ):
                pass
            else:
                raise RuntimeError(f"Unhandled control mode {self.camera_control_mode}")

    def on_zoom(self, scroll: dvec2) -> None:
        if scroll.y == 0:
            return

        if not self.pan_pressed and not self.rotate_pressed:
            self.zoom(scroll, False)

    def on_zoom_with_movement(self, scroll: dvec2) -> None:
        if scroll.y == 0:
            return

        if not self.pan_pressed and not self.rotate_pressed:
            self.zoom(scroll, True)

    def rotate_orbit(self, start_pos: ivec2, pos: ivec2) -> None:
        rot = self._rotation_from_mouse_delta(pos - start_pos, self.camera_target)
        position = vec3(rot * vec4(self.drag_start_camera_position, 1.0))  # type: ignore
        self.camera.camera_from_world = RigidTransform3D.look_at(
            position, self.camera_target, self.camera_world_up, self.handedness
        )

    def rotate_trackball(self, start_pos: ivec2, pos: ivec2) -> None:
        start = self.intersect_trackball(start_pos)
        current = self.intersect_trackball(pos)

        dist = distance(start, current)

        # Skip if starting and current point are too close.
        if dist < 1e-6:
            return

        # Compute axis of rotation as the vector perpendicular to the plane spanned by the
        # vectors connecting the origin to the two points.
        axis = normalize(cross(current, start))

        # Compute angle as the angle between the two vectors, if they are too far away we use the distance
        # between them instead, this makes it continue to rotate when dragging the mouse further away.
        angle = max(acos(dot(normalize(current), normalize(start))), dist)

        # Compute resulting rotation and apply it to the starting position and up vector.
        rot = translate(self.camera_target) * rotate(angle, axis) * translate(-self.camera_target)

        # Update camera
        new_up = vec3(rot * vec4(self.drag_start_camera_up, 0.0))  # type: ignore
        new_camera_position = vec3(rot * vec4(self.drag_start_camera_position, 1.0))  # type: ignore
        self.camera.camera_from_world = RigidTransform3D.look_at(
            new_camera_position, self.camera_target, new_up, self.handedness
        )

    def rotate_first_person(self, start_pos: ivec2, pos: ivec2) -> None:
        rot = self._rotation_from_mouse_delta(pos - start_pos, self.drag_start_camera_position)
        target: vec3 = rot * self.camera_target  # type: ignore
        self.camera.camera_from_world = RigidTransform3D.look_at(
            self.drag_start_camera_position, target, self.camera_world_up, self.handedness
        )

    def pan(self, start_pos: ivec2, pos: ivec2) -> None:
        delta = pos - start_pos
        if isinstance(self.camera, OrthographicCamera):
            delta_inv_y = vec2(delta.x, -delta.y)
            ar_half_extents = self.camera.half_extents * vec2(self.rect.width / self.rect.height, 1)
            movement = delta_inv_y / vec2(self.rect.width, self.rect.height) * ar_half_extents * 2.0
        else:
            speed_scale = max(
                distance(self.camera_target, self.drag_start_camera_position) * self.camera_pan_distance_speed_scale,
                self.camera_pan_min_speed_scale,
            )
            movement = vec2(delta) * self.camera_pan_speed * speed_scale

        delta_position = movement.x * self.drag_start_camera_right + movement.y * self.drag_start_camera_up
        self.camera_target = self.drag_start_camera_target + delta_position
        self.camera.camera_from_world = RigidTransform3D.look_at(
            self.drag_start_camera_position + delta_position, self.camera_target, self.camera_world_up, self.handedness
        )

    def zoom(self, scroll: dvec2, move: bool) -> None:
        if isinstance(self.camera, OrthographicCamera):
            self.camera.half_extents -= self.camera.half_extents * self.camera_zoom_ortho_speed * scroll.y
            # TODO: pan camera to ensure point under cursor stays fixed in screen coordinates
        else:
            position = self.camera.position()
            dist = distance(position, self.camera_target)
            speed_scale = max(dist * self.camera_zoom_distance_speed_scale, self.camera_zoom_min_speed_scale)
            movement = scroll.y * speed_scale * self.camera_zoom_speed
            if move:
                # If moving, also move target
                delta_position = self.camera.front() * movement
                self.camera_target += delta_position
            else:
                # If target not moving, clamp movement to keep a minimum distance to the target
                max_movement = max(dist - self.camera_zoom_min_target_distance, 0.0)
                movement = min(movement, max_movement)
                delta_position = self.camera.front() * movement

            self.camera.camera_from_world = RigidTransform3D.look_at(
                self.camera.position() + delta_position, self.camera_target, self.camera_world_up, self.handedness
            )

    def _rotation_from_mouse_delta(self, delta: ivec2, center_of_rotation: vec3) -> mat4:
        # Clamp pitch algorithm outline:
        # - Take current camera front
        # - Project on XZ plane (assumint Y up) -> we do this by projecting to plane defined by world_up
        # - normalize to project back to distance 1
        # - compute current angle by taking acos of projection to normal
        # - compute new angle by adding speed * y mouse movement to angle
        # - clamp angle to desired value range
        # - rotate by initial camera right with new angle
        #
        # This also works for arbitrary world up, not just axis aligned
        new_pitch = clamp(
            self.drag_start_camera_pitch - delta.y * self.camera_rotation_speed.y, math.pi * 0.01, math.pi * 0.99
        )
        delta_pitch = self.drag_start_camera_pitch - new_pitch
        rot_y = rotate(delta_pitch, self.drag_start_camera_right)
        rot_x = rotate(-self.camera_rotation_speed.x * delta.x, self.camera_world_up)
        return translate(center_of_rotation) * rot_x * rot_y * translate(-center_of_rotation)  # type: ignore

    def intersect_trackball(self, pos: ivec2) -> vec3:
        """
        Return intersection of a line passing through the mouse position at pixel coordinates x, y
        and the trackball as a point in world coordinates.
        """
        width = self.rect.width
        height = self.rect.height

        # Transform mouse coordinates from -1 to 1
        nx = 2 * (pos.x + 0.5) / width - 1
        ny = 1 - 2 * (pos.y + 0.5) / height

        # Adjust coordinates for the larger side of the viewport rectangle.
        if width > height:
            nx *= width / height
        else:
            ny *= height / width

        s = nx * nx + ny * ny
        if s <= 0.5:
            # Sphere intersection
            nz = sqrt(1 - s)
        else:
            # Hyperboloid intersection.
            nz = 1 / (2 * sqrt(s))

        # Return intersection position in world coordinates.
        return self.drag_start_camera_inverse_view_rotation * vec3(nx, ny, nz)  # type: ignore



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from enum import Flag, auto
from typing import TYPE_CHECKING, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import NDArray
from pyxpg import (
    BorderColor,
    BufferUsageFlags,
    DescriptorSetBinding,
    DescriptorType,
    Filter,
    Image,
    ImageLayout,
    ImageUsageFlags,
    ImageView,
    MemoryUsage,
    PipelineStageFlags,
    Sampler,
    SamplerAddressMode,
    SamplerMipmapMode,
    get_format_info,
)

from .property import BufferProperty, ImageProperty, as_buffer_property, as_image_property
from .renderer_frame import RendererFrame
from .utils.descriptors import create_descriptor_layout_pool_and_sets_ringbuffer
from .utils.gpu import UploadableBuffer, align_up, view_bytes
from .utils.ring_buffer import RingBuffer

if TYPE_CHECKING:
    from .renderer import Renderer


class MaterialPropertyFlags(Flag):
    ALLOW_IMAGE = auto()
    HAS_VALUE = auto()
    DEFAULT_SRGB = auto()


@dataclass
class MaterialSamplingOptions:
    min_filter: Filter = Filter.LINEAR
    mag_filter: Filter = Filter.LINEAR
    mipmap_mode: SamplerMipmapMode = SamplerMipmapMode.LINEAR
    mip_lod_bias: float = 0.0
    min_lod: float = 0.0
    max_lod: float = 1000.0
    u: SamplerAddressMode = SamplerAddressMode.CLAMP_TO_EDGE
    v: SamplerAddressMode = SamplerAddressMode.CLAMP_TO_EDGE
    w: SamplerAddressMode = SamplerAddressMode.CLAMP_TO_EDGE
    anisotroy_enabled: bool = False
    max_anisotropy: float = 0.0
    border_color: BorderColor = BorderColor.FLOAT_TRANSPARENT_BLACK


class MaterialProperty:
    def __init__(
        self,
        property: Union[BufferProperty, ImageProperty, float, Tuple[float, ...], NDArray[np.float32]],
        channels: int,
        srgb: bool,
        mips: bool,
        name: str,
    ):
        self.property: Union[BufferProperty, ImageProperty]
        if isinstance(property, ImageProperty):
            img = as_image_property(property, name=name)
            info = get_format_info(img.format)
            if info.channels < channels:
                raise ValueError(
                    f'Material property "{name}" image must have at least {channels}, but {img.format} only has {info.channels}'
                )
            self.property = img
        else:
            buf = as_buffer_property(property, np.float32, (channels,) if channels > 1 else (), name=name)
            self.property = buf
        self.channels = channels
        self.srgb = srgb
        self.mips = mips
        self.name = name


MaterialData = Union[float, Tuple[float, ...], NDArray[np.float32], BufferProperty, ImageProperty, MaterialProperty]


def as_material_property(
    property: MaterialData, expected_channels: int, flags: MaterialPropertyFlags, name: str
) -> Tuple[MaterialProperty, MaterialPropertyFlags]:
    if isinstance(property, MaterialProperty):
        material_property = property
    else:
        material_property = MaterialProperty(
            property, expected_channels, bool(flags & MaterialPropertyFlags.DEFAULT_SRGB), True, name
        )

    if material_property.channels != expected_channels:
        raise ValueError(
            f'Material property "{name}" expects {expected_channels} channels. Got {material_property.channels} channels'
        )

    if isinstance(material_property.property, ImageProperty) and not (flags & MaterialPropertyFlags.ALLOW_IMAGE):
        raise ValueError(f'Material property "{name}" cannot be an image')

    return material_property, flags


class Material:
    def __init__(
        self,
        properties: List[Tuple[MaterialProperty, MaterialPropertyFlags]],
        shader_defines: List[Tuple[str, str]],
        sampling_options: Optional[MaterialSamplingOptions],
    ):
        self.properties = properties
        self.created = False
        self.shader_defines = shader_defines
        self.sampling_options = sampling_options if sampling_options is not None else MaterialSamplingOptions()

        offset = 0
        fields = {}
        for p, flags in properties:
            if flags & MaterialPropertyFlags.HAS_VALUE:
                alignment = p.channels * 4 if p.channels != 3 else 16
                offset = align_up(offset, alignment)
                fields[p.name] = (np.dtype((np.float32, (p.channels,))), offset)
                offset += p.channels * 4

            if flags & MaterialPropertyFlags.ALLOW_IMAGE:
                fields[f"has_{p.name}_texture"] = (np.uint32, offset)  # type: ignore
                offset += 4
        self.dtype = np.dtype(fields)  # type: ignore

        for p, _ in properties:
            if isinstance(p.property, ImageProperty):
                p.property.use_gpu(
                    ImageUsageFlags.TRANSFER_DST
                    | ImageUsageFlags.SAMPLED
                    | ImageUsageFlags.STORAGE
                    | ImageUsageFlags.TRANSFER_SRC,
                    ImageLayout.SHADER_READ_ONLY_OPTIMAL,
                    PipelineStageFlags.FRAGMENT_SHADER,
                    mips=p.mips,
                    srgb=p.srgb,
                )
            p.property.update_callbacks.append(lambda _: self.reupload())
        self.need_upload = True

    def create(self, r: "Renderer") -> None:
        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_sets = (
            create_descriptor_layout_pool_and_sets_ringbuffer(
                r.device,
                [
                    DescriptorSetBinding(1, DescriptorType.UNIFORM_BUFFER),
                    DescriptorSetBinding(1, DescriptorType.SAMPLER),
                ]
                + [
                    DescriptorSetBinding(1, DescriptorType.SAMPLED_IMAGE)
                    for _, flags in self.properties
                    if flags & MaterialPropertyFlags.ALLOW_IMAGE
                ],
                r.num_frames_in_flight,
            )
        )

        self.uniform_buffers = RingBuffer(
            [
                UploadableBuffer(r.device, self.dtype.itemsize, BufferUsageFlags.UNIFORM)
                for _ in range(r.num_frames_in_flight)
            ]
        )

        self.sampler = Sampler(
            r.device,
            self.sampling_options.min_filter,
            self.sampling_options.mag_filter,
            self.sampling_options.mipmap_mode,
            self.sampling_options.mip_lod_bias,
            self.sampling_options.min_lod,
            self.sampling_options.max_lod,
            self.sampling_options.u,
            self.sampling_options.v,
            self.sampling_options.w,
            self.sampling_options.anisotroy_enabled,
            self.sampling_options.max_anisotropy,
            border_color=self.sampling_options.border_color,
        )
        for set, buf in zip(self.descriptor_sets, self.uniform_buffers):
            set.write_buffer(buf, DescriptorType.UNIFORM_BUFFER, 0, 0)
            set.write_sampler(self.sampler, 1, 0)

        self.constants = np.zeros((1,), self.dtype)

    def create_if_needed(self, r: "Renderer") -> None:
        if not self.created:
            self.create(r)
            self.created = True

    def reupload(self) -> None:
        self.need_upload = True

    def upload(self, r: "Renderer", frame: RendererFrame) -> None:
        if not self.need_upload:
            return

        self.descriptor_set = self.descriptor_sets.get_current_and_advance()
        image_index = 0
        for p, flags in self.properties:
            image: Union[Image, ImageView]
            if isinstance(p.property, BufferProperty):
                if flags & MaterialPropertyFlags.HAS_VALUE:
                    self.constants[p.name] = p.property.get_current()
                self.constants[f"has_{p.name}_texture"] = False
                image = r.zero_image
            else:
                if flags & MaterialPropertyFlags.HAS_VALUE:
                    self.constants[p.name] = 0.0
                self.constants[f"has_{p.name}_texture"] = True
                image = p.property.get_current_gpu().view()

            if flags & MaterialPropertyFlags.ALLOW_IMAGE:
                self.descriptor_set.write_image(
                    image, ImageLayout.SHADER_READ_ONLY_OPTIMAL, DescriptorType.SAMPLED_IMAGE, image_index + 2
                )
                image_index += 1

        self.uniform_buffers.get_current_and_advance().upload(
            frame.cmd, MemoryUsage.SHADER_UNIFORM, view_bytes(self.constants)
        )

        self.need_upload = False


# Unlit flat color
class ColorMaterial(Material):
    def __init__(self, color: MaterialData, sampling_options: Optional[MaterialSamplingOptions] = None):
        self.color = as_material_property(
            color, 3, MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE, "color"
        )
        super().__init__([self.color], [("MATERIAL_COLOR", "1")], sampling_options)


# Diffuse only
class DiffuseMaterial(Material):
    def __init__(
        self,
        diffuse: MaterialData,
        normal: Optional[ImageProperty] = None,
        sampling_options: Optional[MaterialSamplingOptions] = None,
    ):
        self.diffuse = as_material_property(
            diffuse,
            3,
            MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE | MaterialPropertyFlags.DEFAULT_SRGB,
            "diffuse",
        )
        self.normal = as_material_property(normal or (0.0, 0.0, 0.0), 3, MaterialPropertyFlags.ALLOW_IMAGE, "normal")
        super().__init__([self.diffuse, self.normal], [("MATERIAL_DIFFUSE", "1")], sampling_options)


# Blinn-phong
class DiffuseSpecularMaterial(Material):
    def __init__(
        self,
        diffuse: MaterialData,
        specular_strength: MaterialData,
        specular_exponent: Union[float, BufferProperty] = 32.0,
        specular_tint: Union[float, BufferProperty] = 0.0,
        normal: Optional[ImageProperty] = None,
        sampling_options: Optional[MaterialSamplingOptions] = None,
    ):
        self.diffuse = as_material_property(
            diffuse,
            3,
            MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE | MaterialPropertyFlags.DEFAULT_SRGB,
            "diffuse",
        )
        self.specular_strength = as_material_property(
            specular_strength,
            1,
            MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE,
            "specular_strength",
        )
        self.specular_exponent = as_material_property(
            specular_exponent, 1, MaterialPropertyFlags.HAS_VALUE, "specular_exponent"
        )
        self.specular_tint = as_material_property(specular_tint, 1, MaterialPropertyFlags.HAS_VALUE, "specular_tint")
        self.normal = as_material_property(normal or (0.0, 0.0, 0.0), 3, MaterialPropertyFlags.ALLOW_IMAGE, "normal")
        super().__init__(
            [self.diffuse, self.specular_strength, self.specular_exponent, self.specular_tint, self.normal],
            [("MATERIAL_DIFFUSE_SPECULAR", "1")],
            sampling_options,
        )


# PBR metallic roughness
class PBRMaterial(Material):
    def __init__(
        self,
        albedo: MaterialData,
        roughness: MaterialData,
        metallic: MaterialData,
        ao: Union[float, BufferProperty] = 1.0,
        normal: Optional[ImageProperty] = None,
        sampling_options: Optional[MaterialSamplingOptions] = None,
    ):
        self.albedo = as_material_property(
            albedo,
            3,
            MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE | MaterialPropertyFlags.DEFAULT_SRGB,
            "albedo",
        )
        self.roughness = as_material_property(
            roughness, 1, MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE, "roughness"
        )
        self.metallic = as_material_property(
            metallic, 1, MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE, "metallic"
        )
        self.ao = as_material_property(
            ao, 1, MaterialPropertyFlags.ALLOW_IMAGE | MaterialPropertyFlags.HAS_VALUE, "ao"
        )
        self.normal = as_material_property(normal or (0.0, 0.0, 0.0), 3, MaterialPropertyFlags.ALLOW_IMAGE, "normal")
        super().__init__(
            [self.albedo, self.roughness, self.metallic, self.ao, self.normal],
            [("MATERIAL_PBR", "1")],
            sampling_options,
        )



import os
import sys
from pathlib import Path

import cv2
import numpy as np
from common import gltf
from pyglm.glm import angleAxis, vec3
from pyxpg import Format

from ambra.config import *
from ambra.geometry import create_sphere
from ambra.lights import EnvironmentLight
from ambra.materials import PBRMaterial
from ambra.primitives3d import Grid, GridType, Mesh
from ambra.property import ArrayImageProperty
from ambra.viewer import Viewer

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"

equirectangular = cv2.imread(sys.argv[2], cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH | cv2.IMREAD_COLOR_RGB)
if equirectangular is None:
    print(f"Equirectangular image load failed: {sys.argv[2]}")
    exit(1)
lights = [EnvironmentLight.from_equirectangular(equirectangular)]

model = gltf.load(Path(sys.argv[1]))

meshes = []
for mesh_data in model.meshes:
    print(mesh_data.ao_texture.shape)
    material = PBRMaterial(
        # albedo=ArrayImageProperty(mesh_data.base_color_texture[np.newaxis], Format.R8G8B8A8_SRGB),
        albedo=ArrayImageProperty(mesh_data.base_color_texture[np.newaxis], Format.R8G8B8A8_UNORM),
        roughness=ArrayImageProperty(mesh_data.metallic_roughness_texture[:, :, 1:2][np.newaxis], Format.R8_UNORM),
        metallic=ArrayImageProperty(mesh_data.metallic_roughness_texture[:, :, 2:3][np.newaxis], Format.R8_UNORM),
        ao=ArrayImageProperty(mesh_data.ao_texture[..., :1][np.newaxis]),
        normal=ArrayImageProperty(mesh_data.normal_texture[np.newaxis]),
    )

    assert np.all(mesh_data.tangents[:, 3] > 0.0)
    m = Mesh(
        mesh_data.positions,
        mesh_data.indices,
        mesh_data.normals,
        mesh_data.tangents[:, :3],
        mesh_data.uvs,
        material=material,
        translation=(0, 0, 2),
        rotation=angleAxis(np.pi * 0.5, vec3(1, 0, 0)),
    )
    meshes.append(m)

mirror = PBRMaterial(
    albedo=(1, 1, 1),
    roughness=(0.01),
    metallic=1,
)
white = PBRMaterial(
    albedo=(1, 1, 1),
    roughness=(1.0),
    metallic=0,
)
dark_gray = PBRMaterial(
    albedo=(0.1, 0.1, 0.1),
    roughness=(1.0),
    metallic=0,
)
light_gray = PBRMaterial(
    albedo=(0.4, 0.4, 0.4),
    roughness=(1.0),
    metallic=0,
)
v, n, f = create_sphere(0.5, rings=32, sectors=64)
spheres = [
    Mesh(v, f, normals=n, translation=(2, 1, 0.0), material=mirror),
    Mesh(v, f, normals=n, translation=(2, 1, 1.1), material=white),
    Mesh(v, f, normals=n, translation=(2, 1, 2.2), material=light_gray),
    Mesh(v, f, normals=n, translation=(2, 1, 3.3), material=dark_gray),
]

g = Grid.transparent_black_lines((100, 100), GridType.XY_PLANE)

v = Viewer(
    config=Config(
        gui=GuiConfig(stats=True, playback=True, inspector=True, renderer=True),
        world_up=(0, 0, 1),
        renderer=RendererConfig(msaa_samples=4),
        camera=CameraConfig(
            position=(4, -4, 3),
            target=(0, 0, 1.5),
            z_near=0.1,
        ),
    )
)

v.scene.objects.extend(lights)
v.scene.objects.extend(meshes)
v.scene.objects.extend(spheres)
v.scene.objects.append(g)

v.run()



import numpy as np
from pyglm.glm import vec3

from ambra.config import CameraConfig, CameraProjection, Config, PlaybackConfig
from ambra.geometry import create_axis2d_lines_and_colors
from ambra.primitives2d import Lines
from ambra.property import AnimationBoundary, ArrayBufferProperty, FrameAnimation
from ambra.transform3d import RigidTransform3D
from ambra.viewer import Viewer

viewer = Viewer(
    config=Config(
        playback=PlaybackConfig(
            playing=True,
        ),
        camera=CameraConfig(
            projection=CameraProjection.ORTHOGRAPHIC,
            ortho_half_extents=(10, 10),
            position=(0, 0, 0),
            target=(0, 0, 1),
        ),
        world_up=(0, 1, 0),
    ),
)

positions, colors = create_axis2d_lines_and_colors()
positions = np.linspace(positions, positions * np.array([2, 1]), 100)
line_width = np.linspace(1, 32, 100)

scale = np.linspace(np.array([1, 1]), np.array([1, 3]), 100)
rotation = ArrayBufferProperty(
    np.linspace(0, 4 * np.pi, 50),
    np.float32,
)
translation = ArrayBufferProperty(
    np.linspace(np.array([0, 0]), np.array([1, 1]), 50),
    np.float32,
    FrameAnimation(boundary=AnimationBoundary.MIRROR),
)

line = Lines(
    positions,
    colors,
    line_width,
    scale=scale,
    translation=translation,
    rotation=rotation,
)

translation = ArrayBufferProperty(
    np.linspace(np.array([1, 0]), np.array([2, 0]), 50),
    np.float32,
    FrameAnimation(boundary=AnimationBoundary.MIRROR),
)
line2 = Lines(positions, colors, line_width, scale=None, translation=translation)

viewer.scene.objects.extend([line, line2])

viewer.run()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, Dict, List

import numpy as np
from pyxpg import DescriptorType, slang


@dataclass
class DescriptorInfo:
    name: str
    binding: int
    set: int
    resource: slang.Resource
    image_format: slang.ImageFormat
    count: int = 1  # 1 for single element, 0 for unbounded


def to_descriptor_type(binding_type: slang.BindingType) -> DescriptorType:
    if binding_type == slang.BindingType.SAMPLER:
        return DescriptorType.SAMPLER
    elif binding_type == slang.BindingType.COMBINED_TEXTURE_SAMPLER:
        return DescriptorType.COMBINED_IMAGE_SAMPLER
    elif binding_type == slang.BindingType.TEXTURE:
        return DescriptorType.SAMPLED_IMAGE
    elif binding_type == slang.BindingType.MUTABLE_TEXTURE:
        return DescriptorType.STORAGE_IMAGE
    elif binding_type == slang.BindingType.TYPED_BUFFER:
        return DescriptorType.UNIFORM_TEXEL_BUFFER
    elif binding_type == slang.BindingType.MUTABLE_TYPED_BUFFER:
        return DescriptorType.STORAGE_TEXEL_BUFFER
    elif binding_type == slang.BindingType.RAW_BUFFER:
        return DescriptorType.STORAGE_BUFFER
    elif binding_type == slang.BindingType.MUTABLE_RAW_BUFFER:
        return DescriptorType.STORAGE_BUFFER
    elif binding_type == slang.BindingType.INPUT_RENDER_TARGET:
        return DescriptorType.INPUT_ATTACHMENT
    elif binding_type == slang.BindingType.INLINE_UNIFORM_DATA:
        return DescriptorType.INLINE_UNIFORM_BLOCK
    elif binding_type == slang.BindingType.RAYTRACING_ACCELERATION_STRUCTURE:
        return DescriptorType.ACCELERATION_STRUCTURE
    elif binding_type == slang.BindingType.CONSTANT_BUFFER:
        return DescriptorType.UNIFORM_BUFFER
    else:
        raise ValueError("Unknown slang.BindingType {}", binding_type)


def _flatten_descriptor_sets(
    sets: Dict[int, List[DescriptorInfo]],
    typ: slang.Type,
    name: str,
    binding: int,
    set: int,
    image_format: slang.ImageFormat,
) -> None:
    if isinstance(typ, (slang.Scalar, slang.Vector, slang.Matrix)):
        pass
    elif isinstance(typ, slang.Array):
        child_typ = typ.type
        if isinstance(child_typ, slang.Resource):
            sets.setdefault(set, []).append(DescriptorInfo(name, binding, set, child_typ, image_format, typ.count))
    elif isinstance(typ, slang.Struct):
        for f in typ.fields:
            _flatten_descriptor_sets(
                sets,
                f.type,
                f"{name}.{f.name}" if name else f.name,
                binding + f.binding,
                set + f.set,
                f.image_format,
            )
    elif isinstance(typ, slang.Resource):
        sets.setdefault(set, []).append(DescriptorInfo(name, binding, set, typ, image_format))
    else:
        raise TypeError(f"Unknown slang.Type: {type(typ)}")


class DescriptorSetsReflection:
    def __init__(self, refl: slang.Reflection):
        self.sets: Dict[int, List[DescriptorInfo]] = {}
        _flatten_descriptor_sets(self.sets, refl.object, "", 0, 0, slang.ImageFormat.UNKNOWN)

        self.descriptors: Dict[str, DescriptorInfo] = {}
        for s in self.sets.values():
            for r in s:
                self.descriptors[r.name] = r


_scalar_to_np = {
    slang.ScalarKind.BOOL: np.dtype(bool),
    slang.ScalarKind.INT8: np.dtype(np.int8),
    slang.ScalarKind.UINT8: np.dtype(np.uint8),
    slang.ScalarKind.INT16: np.dtype(np.int16),
    slang.ScalarKind.UINT16: np.dtype(np.uint16),
    slang.ScalarKind.INT32: np.dtype(np.int32),
    slang.ScalarKind.UINT32: np.dtype(np.uint32),
    slang.ScalarKind.INT64: np.dtype(np.int64),
    slang.ScalarKind.UINT64: np.dtype(np.uint64),
    slang.ScalarKind.FLOAT16: np.dtype(np.float16),
    slang.ScalarKind.FLOAT32: np.dtype(np.float32),
    slang.ScalarKind.FLOAT64: np.dtype(np.float64),
}


def to_dtype(typ: slang.Type) -> np.dtype[Any]:
    if isinstance(typ, slang.Scalar):
        return _scalar_to_np[typ.base]
    elif isinstance(typ, slang.Vector):
        return np.dtype((_scalar_to_np[typ.base], (typ.count,)))
    elif isinstance(typ, slang.Matrix):
        return np.dtype((_scalar_to_np[typ.base], (typ.rows, typ.columns)))
    elif isinstance(typ, slang.Array):
        return np.dtype((to_dtype(typ.type), (typ.count,)))
    elif isinstance(typ, slang.Struct):
        d = {}
        for f in typ.fields:
            d[f.name] = (to_dtype(f.type), f.offset)
        return np.dtype(d)  # type: ignore
    else:
        raise TypeError(f"Unknown slang.Type: {type(typ)}")


def print_type(typ: slang.Type, indent: int = 0) -> None:
    if isinstance(typ, slang.Scalar):
        print(" " * indent + f"{typ.base}")  # noqa: T201
    elif isinstance(typ, slang.Vector):
        print(" " * indent + f"{typ.base}_{typ.count}")  # noqa: T201
    elif isinstance(typ, slang.Matrix):
        print(" " * indent + f"{typ.base}_{typ.rows}x{typ.columns}")  # noqa: T201
    elif isinstance(typ, slang.Array):
        child_typ = typ.type
        if isinstance(child_typ, slang.Resource):
            print(" " * indent + f"{child_typ.kind}[{typ.count}]")  # noqa: T201
        else:
            print(" " * indent + f"{typ.type}[{typ.count}]")  # noqa: T201
    elif isinstance(typ, slang.Struct):
        print(" " * indent + "Struct")  # noqa: T201
        for f in typ.fields:
            print(" " * (indent + 4) + f"{f.offset:3d} ({f.binding}, {f.set}) | {f.image_format} | {f.name}: ", end="")  # noqa: T201
            print_type(f.type, indent + 4)
    elif isinstance(typ, slang.Resource):
        print(" " * indent + f"Resource[{typ.kind}]")  # noqa: T201
        if typ.kind == slang.ResourceKind.SAMPLER:
            pass
        elif typ.kind == slang.ResourceKind.TEXTURE_2D:
            pass
        elif typ.kind == slang.ResourceKind.STRUCTURED_BUFFER:
            print_type(typ.type, indent + 4)
        elif typ.kind == slang.ResourceKind.CONSTANT_BUFFER:
            print_type(typ.type, indent + 4)
    else:
        print(" " * indent + f"Unknown slang.Type: {type(typ)}")  # noqa: T201



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import asyncio
import atexit
import json
import logging
import pickle
import struct
from dataclasses import dataclass
from enum import Enum
from threading import Thread
from typing import Callable, Dict, Optional, Tuple, Union

from .config import ServerConfig

logger = logging.getLogger(__name__)


@dataclass
class Client:
    address: str
    port: int
    name: str


class MessageId(Enum):
    NONE = 0  # Reserved for end of connection

    # Default message ids
    INPUT = 1
    OBJECT = 2

    # Starting id available for users
    USER = 1 << 20


class Format(Enum):
    # Default formats
    BINARY = 0
    JSON = 1
    PICKLE = 2

    # Starting id available for users
    USER = 1 << 20


@dataclass
class RawMessage:
    id: int
    format: int
    data: bytes


class Server:
    def __init__(
        self,
        on_raw_message_async: Callable[[Client, RawMessage], None],
        config: ServerConfig,
    ):
        self.connections: Dict[Tuple[str, int], asyncio.StreamWriter] = {}

        def entry() -> None:
            async def handle(reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
                client_name = ""
                client_address = ("", 0)

                try:
                    client_address = writer.get_extra_info("peername")
                    self.connections[client_address] = writer

                    logger.info("Server: client %s:%s opened connection", client_address[0], client_address[1])
                    magic = await reader.readexactly(4)
                    if magic != b"AMBR":
                        raise ValueError("invalid magic")  # noqa: TRY301

                    # Header
                    client_name_length = struct.unpack("<I", await reader.readexactly(4))[0]
                    if client_name_length > 0:
                        client_name = (await reader.readexactly(client_name_length)).decode(
                            encoding="utf-8", errors="ignore"
                        )
                    client = Client(client_address[0], client_address[1], client_name)

                    logger.info(
                        'Server: client %s:%s registered as "%s"', client_address[0], client_address[1], client_name
                    )

                    # Messages
                    while True:
                        id = struct.unpack("<I", await reader.readexactly(4))[0]
                        if id == 0:
                            logger.info(
                                "Server: client %s(%s:%s) closed connection",
                                client_address[0],
                                client_address[1],
                                client_name,
                            )
                            break
                        format, length = struct.unpack("<IQ", await reader.readexactly(12))
                        data = await reader.readexactly(length)

                        on_raw_message_async(client, RawMessage(id, format, data))
                except asyncio.exceptions.IncompleteReadError:
                    logger.info(
                        "Server: client %s(%s:%s) unexpected EOF before closing connection",
                        client_name,
                        client_address[0],
                        client_address[1],
                    )
                except Exception as e:
                    logger.info(
                        "Server: exception while handling client connection %s(%s:%s): %s",
                        client_name,
                        client_address[0],
                        client_address[1],
                        e,
                    )

                del self.connections[client_address]
                writer.close()
                await writer.wait_closed()

            async def main() -> None:
                server = await asyncio.start_server(handle, config.address, config.port)

                async with server:
                    await server.start_serving()
                    await self.stop

                    # Stop accepting new connections
                    server.close()

                    # Close all existing open connections
                    connections = self.connections.values()
                    for c in connections:
                        c.close()

            asyncio.set_event_loop(self.loop)
            self.loop.run_until_complete(main())

        self.enabled = config.enabled

        if self.enabled:
            self.loop = asyncio.new_event_loop()
            self.stop = self.loop.create_future()

            self.thread = Thread(None, entry, "Server", daemon=True)
            self.thread.start()
            atexit.register(self.shutdown)

    def shutdown(self) -> None:
        if self.enabled and not self.stop.done():
            self.loop.call_soon_threadsafe(self.stop.set_result, None)
            self.thread.join()


@dataclass
class InputMessage:
    num: int

    @classmethod
    def from_raw(cls, data: bytes) -> "InputMessage":
        return InputMessage(0)

    @classmethod
    def from_json(cls, obj: object) -> "InputMessage":
        return InputMessage(0)


@dataclass
class ObjectMessage:
    @classmethod
    def from_raw(cls, data: bytes) -> "ObjectMessage":
        return ObjectMessage()

    @classmethod
    def from_json(cls, obj: object) -> "ObjectMessage":
        return ObjectMessage()


Message = Union[
    InputMessage,
    ObjectMessage,
]

_binary_dispatch: Dict[int, Callable[[bytes], Message]] = {
    MessageId.INPUT.value: InputMessage.from_raw,
    MessageId.OBJECT.value: ObjectMessage.from_raw,
}

_json_dispatch: Dict[int, Callable[[object], Message]] = {
    MessageId.INPUT.value: InputMessage.from_json,
    MessageId.OBJECT.value: ObjectMessage.from_json,
}


def _parse_builtin_messages_binary(raw: RawMessage) -> Optional[Message]:
    fn = _binary_dispatch.get(raw.id)
    if not fn:
        return None
    return fn(raw.data)


def _parse_builtin_messages_json(raw: RawMessage) -> Optional[Message]:
    fn = _json_dispatch.get(raw.id)
    if not fn:
        return None
    obj = json.loads(raw.data)
    return fn(obj)


def _parse_builtin_messages_pickle(raw: RawMessage) -> Optional[Message]:
    return pickle.loads(raw.data)  # type: ignore


_dispatch = {
    Format.BINARY.value: _parse_builtin_messages_binary,
    Format.JSON.value: _parse_builtin_messages_json,
    Format.PICKLE.value: _parse_builtin_messages_pickle,
}


def parse_builtin_messages(raw: RawMessage) -> Optional[Message]:
    format_fn = _dispatch.get(raw.format)
    if not format_fn:
        return None

    return format_fn(raw)



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Dict, List, Optional

from pyxpg import (
    BufferBarrier,
    CommandBuffer,
    ImageBarrier,
    PipelineStageFlags,
    TimelineSemaphore,
)

from .ffx import MipGenerationRequest
from .utils.gpu import MipGenerationFilter


@dataclass
class SemaphoreInfo:
    sem: TimelineSemaphore
    wait_stage: PipelineStageFlags
    wait_value: int
    signal_stage: PipelineStageFlags
    signal_value: int


@dataclass
class RendererFrame:
    index: int
    total_index: int
    path_tracing: bool

    cmd: CommandBuffer

    # Before and after upload barriers
    upload_property_pipeline_stages: PipelineStageFlags
    upload_before_buffer_barriers: List[BufferBarrier]
    upload_before_image_barriers: List[ImageBarrier]
    upload_after_buffer_barriers: List[BufferBarrier]
    upload_after_image_barriers: List[ImageBarrier]

    # Mip generation requests
    mip_generation_requests: Dict[MipGenerationFilter, List[MipGenerationRequest]]

    before_render_src_pipeline_stages: PipelineStageFlags
    before_render_dst_pipeline_stages: PipelineStageFlags
    before_render_image_barriers: List[ImageBarrier]

    between_viewport_render_src_pipeline_stages: PipelineStageFlags
    between_viewport_render_dst_pipeline_stages: PipelineStageFlags

    additional_semaphores: List[SemaphoreInfo]

    transfer_cmd: Optional[CommandBuffer]
    transfer_semaphores: List[SemaphoreInfo]

    transfer_upload_before_buffer_barriers: List[BufferBarrier]
    transfer_upload_before_image_barriers: List[ImageBarrier]
    transfer_upload_after_buffer_barriers: List[BufferBarrier]
    transfer_upload_after_image_barriers: List[ImageBarrier]



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING, Callable, List, Optional, Set, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, DTypeLike
from pyglm.glm import mat3, mat4, quat, vec2, vec3
from pyxpg import DescriptorSet, imgui

from .property import (
    BufferProperty,
    ImageProperty,
    Property,
    as_buffer_property,
    as_image_property,
)
from .renderer_frame import RendererFrame
from .transform2d import Transform2D
from .transform3d import Transform3D
from .utils.gpu import AccelerationStructureInstanceInfo

if TYPE_CHECKING:
    from .materials import Material
    from .renderer import Renderer
    from .viewport import Viewport


_counter = 0


def next_id() -> int:
    global _counter
    _counter += 1
    return _counter


class Object:
    def __init__(
        self,
        name: Optional[str] = None,
        material: Optional["Material"] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        self.uid = next_id()
        self.name = name or f"{type(self).__name__}<{self.uid}>"
        self.children: List[Object] = []
        self.properties: List[Property] = []
        self.material = material
        self.viewport_mask = viewport_mask

        self.created = False
        self.gui_enabled = True
        self.gui_expanded = True
        self.gui_selected = False
        self.gui_selected_property: Optional[Property] = None

        self.enabled = self.add_buffer_property(enabled if enabled is not None else True, bool, name="enabled")

    def add_buffer_property(
        self,
        prop: Union[BufferProperty, List[ArrayLike], ArrayLike],
        dtype: Optional[DTypeLike] = None,
        shape: Tuple[int, ...] = (),
        name: str = "",
    ) -> BufferProperty:
        property = as_buffer_property(prop, dtype, shape, name)
        self.properties.append(property)
        return property

    def add_image_property(
        self,
        prop: Union[ImageProperty, List[ArrayLike], ArrayLike],
        is_3d: bool = False,
        name: str = "",
    ) -> ImageProperty:
        property = as_image_property(prop, is_3d, name=name)
        self.properties.append(property)
        return property

    def create_if_needed(self, renderer: "Renderer") -> None:
        if self.material is not None:
            self.material.create_if_needed(renderer)

        if not self.created:
            self.create(renderer)
            self.created = True

    def create(self, renderer: "Renderer") -> None:
        pass

    def collect_dynamic_properties(self, all_properties: Set[Property]) -> None:
        for p in self.properties:
            if p.is_dynamic():
                all_properties.add(p)
        if self.material is not None:
            for mp, _ in self.material.properties:
                if mp.property.is_dynamic():
                    all_properties.add(mp.property)

    def update(self, time: float, frame: int) -> None:
        pass

    def update_transform(self, parent: Optional["Object"]) -> None:
        # TODO: can merge with udpate? Where to find parent? With link?
        pass

    def upload(self, renderer: "Renderer", frame: RendererFrame) -> None:
        pass

    def post_upload(self, renderer: "Renderer", frame: RendererFrame) -> None:
        pass

    def append_acceleration_structure_instances(
        self, instances: List[AccelerationStructureInstanceInfo], material_index: int
    ) -> None:
        pass

    def pre_render(
        self, renderer: "Renderer", frame: RendererFrame, viewport: "Viewport", scene_descriptor_set: DescriptorSet
    ) -> None:
        pass

    def render(
        self, renderer: "Renderer", frame: RendererFrame, viewport: "Viewport", scene_descriptor_set: DescriptorSet
    ) -> None:
        pass

    def render_depth(self, renderer: "Renderer", frame: RendererFrame, scene_descriptor_set: DescriptorSet) -> None:
        pass

    def destroy(self) -> None:
        for p in self.properties:
            p.destroy()

    def gui(self) -> None:
        imgui.text("Properties:")
        imgui.indent(5)
        for p in self.properties:
            s = imgui.selectable(f"{p.name}", self.gui_selected_property == p)
            if s:
                self.gui_selected_property = p
            if self.gui_selected_property == p:
                imgui.indent(10)

                if isinstance(p, BufferProperty):
                    imgui.text(f"{p.shape} {np.dtype(p.dtype).name} {p.max_size}")
                elif isinstance(p, ImageProperty):
                    imgui.text(f"[{p.width}x{p.height}] {p.format}")

                imgui.text(f"{p.current_frame_index} / {p.num_frames}")

                imgui.text(f"Enabled: {p.current_animation_enabled}")
                imgui.text(f"Dynamic: {p.is_dynamic()}")
                imgui.indent(-10)
        imgui.indent(-5)


class Object2D(Object):
    def __init__(
        self,
        name: Optional[str],
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        super().__init__(name, enabled=enabled, viewport_mask=viewport_mask)
        self.translation = self.add_buffer_property(
            translation if translation is not None else np.array([0, 0]),
            np.float32,
            (2,),
            name="translation",
        )
        self.rotation = self.add_buffer_property(
            rotation if rotation is not None else np.array([0]),
            np.float32,
            name="rotation",
        )
        self.scale = self.add_buffer_property(
            scale if scale is not None else np.array([1, 1]),
            np.float32,
            (2,),
            name="scale",
        )
        self.update_transform(None)

    def update_transform(self, parent: Optional[Object]) -> None:
        assert parent is None or isinstance(parent, Object2D)

        self.current_relative_transform = Transform2D(
            vec2(self.translation.get_current()),
            self.rotation.get_current().item(),
            vec2(self.scale.get_current()),
        )
        self.current_transform_matrix: mat3 = (
            parent.current_transform_matrix if parent is not None else np.eye(3)
        ) @ self.current_relative_transform.as_mat3()  # type: ignore


class Object3D(Object):
    def __init__(
        self,
        name: Optional[str] = None,
        translation: Optional[BufferProperty] = None,
        rotation: Optional[BufferProperty] = None,
        scale: Optional[BufferProperty] = None,
        material: Optional["Material"] = None,
        enabled: Optional[BufferProperty] = None,
        viewport_mask: Optional[int] = None,
    ):
        super().__init__(name, material, enabled, viewport_mask)
        self.translation = self.add_buffer_property(
            translation if translation is not None else np.array([0, 0, 0]),
            np.float32,
            (3,),
            name="translation",
        )
        self.rotation = self.add_buffer_property(
            rotation if rotation is not None else np.array([1, 0, 0, 0]),
            np.float32,
            (4,),
            name="rotation",
        )
        self.scale = self.add_buffer_property(
            scale if scale is not None else np.array([1, 1, 1]),
            np.float32,
            (3,),
            name="scale",
        )
        self.update_transform(None)

    def update_transform(self, parent: Optional[Object]) -> None:
        assert parent is None or isinstance(parent, Object3D)

        self.current_relative_transform = Transform3D(
            vec3(self.translation.get_current()),
            quat(self.rotation.get_current()),
            vec3(self.scale.get_current()),
        )
        self.current_transform_matrix: mat4 = (
            parent.current_transform_matrix if parent is not None else mat4(1.0)
        ) * self.current_relative_transform.as_mat4()  # type: ignore


class Widget:
    def __init__(self, title: str):
        self.title = title
        self.uid = next_id()

        self.created = False
        self.properties: List[Property] = []

    def add_buffer_property(
        self,
        prop: Union[BufferProperty, List[ArrayLike], ArrayLike],
        dtype: Optional[DTypeLike] = None,
        shape: Tuple[int, ...] = (),
        name: str = "",
    ) -> BufferProperty:
        property = as_buffer_property(prop, dtype, shape, name)
        self.properties.append(property)
        return property

    def add_image_property(
        self,
        prop: Union[ImageProperty, List[ArrayLike], ArrayLike],
        is_3d: bool = False,
        name: str = "",
    ) -> ImageProperty:
        property = as_image_property(prop, is_3d, name=name)
        self.properties.append(property)
        return property

    def collect_dynamic_properties(self, all_properties: Set[Property]) -> None:
        for p in self.properties:
            if p.is_dynamic():
                all_properties.add(p)

    def create_if_needed(self, renderer: "Renderer") -> None:
        if not self.created:
            self.create(renderer)
            self.created = True

    def create(self, r: "Renderer") -> None:
        pass

    def upload(self, r: "Renderer", frame: RendererFrame) -> None:
        pass

    def gui(self) -> None:
        pass


class Scene:
    def __init__(self, name: str):
        self.name = name
        self.objects: List[Object] = []
        self.widgets: List[Widget] = []

    def visit_objects_with_parent(self, function: Callable[[Optional[Object], Object], None]) -> None:
        def visit_recursive(p: Optional[Object], o: Object) -> None:
            function(p, o)
            for c in o.children:
                visit_recursive(o, c)

        for o in self.objects:
            visit_recursive(None, o)

    def visit_objects(self, function: Callable[[Object], None]) -> None:
        def visit_recursive(o: Object) -> None:
            function(o)
            for c in o.children:
                visit_recursive(c)

        for o in self.objects:
            visit_recursive(o)

    def visit_objects_pre_post(self, pre: Callable[[Object], bool], post: Callable[[Object], None]) -> None:
        def visit_recursive(o: Object) -> None:
            rec = pre(o)
            if rec:
                for c in o.children:
                    visit_recursive(c)
                post(o)

        for o in self.objects:
            visit_recursive(o)

    def end_animation_time(self, frames_per_second: float) -> float:
        time = 0.0

        all_dynamic_properties: Set[Property] = set()

        def visit(o: Object) -> None:
            o.collect_dynamic_properties(all_dynamic_properties)

        self.visit_objects(visit)
        for w in self.widgets:
            w.collect_dynamic_properties(all_dynamic_properties)

        for p in all_dynamic_properties:
            t = p.end_animation_time(frames_per_second)
            time = max(time, t)

        return time

    def update(self, time: float, frame: int) -> List[Object]:
        all_dynamic_properties: Set[Property] = set()

        def collect(o: Object) -> None:
            o.collect_dynamic_properties(all_dynamic_properties)

        self.visit_objects(collect)
        for w in self.widgets:
            w.collect_dynamic_properties(all_dynamic_properties)

        for p in all_dynamic_properties:
            p.update(time, frame)

        enabled_objects: List[Object] = []

        def update(p: Optional[Object], o: Object) -> None:
            o.update_transform(p)
            if o.enabled.get_current():
                enabled_objects.append(o)

        self.visit_objects_with_parent(update)
        return enabled_objects
